"""Pytest fixtures and settings used in the qnexus integration tests."""

import logging
from contextlib import contextmanager
from pathlib import Path
from typing import Callable, ContextManager, Generator, Literal, Union, cast

import pandas as pd
import pytest
from hugr.package import Package
from pytket.circuit import Circuit
from pytket.qir import pytket_to_qir  # type: ignore[attr-defined]
from pytket.wasm.wasm import WasmFileHandler, WasmModuleHandler
from quantinuum_schemas.models.backend_config import (
    AerConfig,
    AerStateConfig,
    AerUnitaryConfig,
    BackendConfig,
    BasicEmulatorConfig,
    BraketConfig,
    IBMQConfig,
    IBMQEmulatorConfig,
    QuantinuumConfig,
    QulacsConfig,
    StandardEmulatorConfig,
)

import qnexus as qnx
from qnexus.client.auth import login_no_interaction
from qnexus.config import CONFIG
from qnexus.exceptions import (
    NoUniqueMatch,
    ResourceDeleteFailed,
    ResourceUpdateFailed,
    ZeroMatches,
)
from qnexus.filesystem import load, save
from qnexus.models.references import (
    CircuitRef,
    CompileJobRef,
    ExecuteJobRef,
    GpuDecoderConfigRef,
    HUGRRef,
    ProjectRef,
    QIRRef,
    Ref,
    TeamRef,
    WasmModuleRef,
)

test_run_identifier = ""

AllBackendConfigs = Union[
    AerConfig,
    AerStateConfig,
    AerUnitaryConfig,
    BackendConfig,
    BasicEmulatorConfig,
    BraketConfig,
    IBMQConfig,
    IBMQEmulatorConfig,
    QuantinuumConfig,
    QulacsConfig,
    StandardEmulatorConfig,
]


LOGGER = logging.getLogger(__name__)


def pytest_addoption(parser: pytest.Parser) -> None:
    parser.addoption(
        "--purge-projects",
        action="store_true",
        default=False,
        help="Purge the projects created by the tests.",
    )

    parser.addoption(
        "--run-id",
        action="store",
        default="",
        help="Use this to identify resources created by the tests. If not provided the `testrun_uid` from pytest-xdist is used.",
    )


@pytest.fixture(autouse=True)
def set_test_run_identifier(request: pytest.FixtureRequest, testrun_uid: str) -> None:
    global test_run_identifier

    run_id_option = request.config.getoption("--run-id")
    test_run_identifier = run_id_option if run_id_option else testrun_uid


@pytest.fixture(name="test_case_name")
def fixture_test_case_name(request: pytest.FixtureRequest) -> str:
    test_case_name = f"{request.node.module.__name__}::{request.node.name}"
    return f"qnexus.{test_case_name}, run:{request.node.execution_count}, id:{test_run_identifier}"


@pytest.fixture(name="test_suite_name")
def fixture_test_suite_name(request: pytest.FixtureRequest) -> str:
    test_suite_name = f"{request.node.module.__name__}"
    return f"qnexus.{test_suite_name}, run:{request.node.execution_count}, id:{test_run_identifier}"


@pytest.fixture(scope="session", autouse=True)
def authenticated_nexus(
    user_email: str = CONFIG.qa_user_email,
    user_password: str = CONFIG.qa_user_password,
) -> Generator[None, None, None]:
    """Authenticated nexus instance fixture."""
    try:
        login_no_interaction(user_email, user_password)
        yield
    finally:
        qnx.auth.logout()


@pytest.fixture(name="create_team")
def fixture_create_team() -> Callable[[str], ContextManager[TeamRef]]:
    """Returns a `ContextManager` that yields a `TeamRef`
    of an existing/created project with the specified name."""

    @contextmanager
    def make_team_if_needed(
        team_name: str,
    ) -> Generator[TeamRef, None, None]:
        team_ref = None
        try:
            team_ref = qnx.teams.get(team_name)
        except ZeroMatches:
            team_ref = qnx.teams.create(
                name=team_name, description=f"description for {team_name}"
            )
        except Exception as e:
            raise e

        yield team_ref

    return make_team_if_needed


@pytest.fixture(name="create_project")
def fixture_create_project(
    request: pytest.FixtureRequest,
) -> Callable[[str], ContextManager[ProjectRef]]:
    """Returns a `ContextManager` that yields a `ProjectRef` of the
    project created with the specified name."""

    @contextmanager
    def make_temp_project(
        project_name: str,
    ) -> Generator[ProjectRef, None, None]:
        my_proj = qnx.projects.get_or_create(
            name=project_name, description=f"description for {project_name}"
        )
        yield my_proj

        if request.config.getoption("--purge-projects"):
            # If fixtures that use the `create_project`fixture are chained or
            # composed, multiple calls to archive and delete the same project
            # will be executed. In case that happens, we are just logging
            # a warning here.
            try:
                qnx.projects.update(my_proj, archive=True)
                qnx.projects.delete(my_proj)
            except ResourceUpdateFailed:
                LOGGER.warning(f"Project {project_name} is already archived")
            except ResourceDeleteFailed:
                LOGGER.warning(f"Project {project_name} is already deleted")
            except Exception as e:
                raise e

    return make_temp_project


@pytest.fixture(name="create_circuit_in_project")
def fixture_create_circuit_in_project(
    authenticated_nexus: None,
) -> Callable[[Circuit, str, str], ContextManager[CircuitRef]]:
    """Returns a `ContextManager` that yields a `CircuitRef` of the
    specified `circuit` uploaded to a project with `project_name`."""

    @contextmanager
    def upload_circuit_if_needed(
        circuit: Circuit,
        project_name: str,
        circuit_name: str,
    ) -> Generator[CircuitRef, None, None]:
        circuit_ref = _get_or_create_circuit(
            circuit=circuit, project_name=project_name, circuit_name=circuit_name
        )

        yield circuit_ref

    return upload_circuit_if_needed


@pytest.fixture(name="create_compile_job_in_project")
def fixture_create_compile_job_in_project(
    create_circuit_in_project: Callable[
        [Circuit, str, str], ContextManager[CircuitRef]
    ],
) -> Callable[..., ContextManager[CompileJobRef]]:
    """Returns a `ContextManager` that yields a `CompileJobRef` of the
    compile job submitted with the specified `CircuitRef` to the project
    with the provided name.

    Notes:
        - The default backend_config is `qnx.AerConfig()`
        - By default `skip_intermediate_circuits` is False

    Warning:
        - Will create the project and upload the circuit if needed.
        - Will not wait for the job to finish before yielding."""

    @contextmanager
    def make_compile_job(
        project_name: str,
        job_name: str,
        circuit: Circuit,
        circuit_name: str,
        backend_config: AllBackendConfigs | None = None,
        skip_intermediate_circuits: bool = False,
    ) -> Generator[CompileJobRef, None, None]:
        with create_circuit_in_project(
            circuit,
            project_name,
            circuit_name,
        ) as circuit_ref:
            config = backend_config if backend_config is not None else qnx.AerConfig()

            try:
                compile_job_ref = qnx.jobs.get(name_like=job_name)
            except ZeroMatches:
                proj_ref = qnx.projects.get(name_like=project_name)

                compile_job_ref = qnx.start_compile_job(
                    programs=[circuit_ref],
                    name=job_name,
                    description=f"description for {job_name}",
                    project=proj_ref,
                    backend_config=config,
                    skip_intermediate_circuits=skip_intermediate_circuits,
                )

            except Exception as e:
                raise e

            assert isinstance(compile_job_ref, CompileJobRef)
            yield compile_job_ref

    return make_compile_job


@pytest.fixture(name="create_execute_job_in_project")
def fixture_create_execute_job_in_project(
    create_circuit_in_project: Callable[
        [Circuit, str, str], ContextManager[CircuitRef]
    ],
) -> Callable[..., ContextManager[ExecuteJobRef]]:
    """Returns a `ContextManager` that yields a `ExecuteJobRef` of the
    execute job submitted with the specified `CircuitRef` to the project
    with the provided name.

    Notes:
        - A single element/circuit job is created.
        - The default backend_config is `qnx.AerConfig()`.
        - By default 10 shots are executed.

    Warning:
        - Will create the project and upload the circuit if needed.
        - Will not wait for the job to finish before yielding."""

    @contextmanager
    def make_execute_job(
        project_name: str,
        job_name: str,
        circuit: Circuit,
        circuit_name: str,
        backend_config: AllBackendConfigs | None = None,
        n_shots: int = 10,
    ) -> Generator[ExecuteJobRef, None, None]:
        with create_circuit_in_project(
            circuit,
            project_name,
            circuit_name,
        ) as circuit_ref:
            config = backend_config if backend_config is not None else qnx.AerConfig()

            try:
                execute_job_ref = qnx.jobs.get(name_like=job_name)
            except (ZeroMatches, NoUniqueMatch):
                proj_ref = qnx.projects.get(name_like=project_name)

                execute_job_ref = qnx.start_execute_job(
                    programs=[circuit_ref],
                    name=job_name,
                    description=f"description for {job_name}",
                    project=proj_ref,
                    backend_config=config,
                    n_shots=[n_shots],
                )

            assert isinstance(execute_job_ref, ExecuteJobRef)
            yield execute_job_ref

    return make_execute_job


@pytest.fixture(name="create_property_in_project")
def fixture_create_property_in_project(
    create_project: Callable[[str], ContextManager[ProjectRef]],
) -> Callable[..., ContextManager[ProjectRef]]:
    """Returns a `ContextManager` that yields a `ProjectRef` to the project with
    `project_name` where the specified property was created.

    Notes: will create the project if it does not exist."""

    @contextmanager
    def make_temp_property(
        project_name: str,
        property_name: str,
        property_type: Literal["bool", "int", "float", "string"],
        required: bool = False,
    ) -> Generator[ProjectRef, None, None]:
        with create_project(project_name) as proj_ref:
            qnx.projects.add_property(
                name=property_name,
                property_type=property_type,
                project=proj_ref,
                required=required,
            )

            yield proj_ref

    return make_temp_property


@pytest.fixture(name="create_qir_in_project")
def fixture_create_qir_in_project(
    create_project: Callable[[str], ContextManager[ProjectRef]],
) -> Callable[[str, str, bytes], ContextManager[QIRRef]]:
    """Returns a `ContextManager` that yields a `QIRRef` to the `qir_bitcode`
    uploaded to a project with `project_name`.

    Notes: will create the project if it does not exist."""

    @contextmanager
    def upload_qir(
        project_name: str,
        qir_name: str,
        qir: bytes,
    ) -> Generator[QIRRef, None, None]:
        with create_project(project_name) as proj_ref:
            qir_ref = qnx.qir.upload(qir=qir, name=qir_name, project=proj_ref)

            yield qir_ref

    return upload_qir


@pytest.fixture(name="create_hugr_in_project")
def fixture_create_hugr_in_project(
    create_project: Callable[[str], ContextManager[ProjectRef]],
) -> Callable[[str, str, Package], ContextManager[HUGRRef]]:
    """Returns a `ContextManager` that yields a `HUGRRef` to the `hugr_package`
    uploaded to a project with `project_name`.

    Notes: will create the project if it does not exist."""

    @contextmanager
    def upload_hugr(
        project_name: str,
        hugr_name: str,
        hugr_package: Package,
    ) -> Generator[HUGRRef, None, None]:
        with create_project(project_name) as proj_ref:
            hugr_ref = qnx.hugr.upload(
                hugr_package=hugr_package,
                name=hugr_name,
                project=proj_ref,
            )

            yield hugr_ref

    return upload_hugr


@pytest.fixture(name="create_wasm_in_project")
def fixture_create_wasm_in_project(
    create_project: Callable[[str], ContextManager[ProjectRef]],
) -> Callable[[str, str, WasmModuleHandler], ContextManager[WasmModuleRef]]:
    """Returns a `ContextManager` that yields a `WasmModuleRef` to the `wasm_module`
    uploaded to a project with `project_name`.

    Notes: will create the project if it does not exist."""

    @contextmanager
    def upload_wasm(
        project_name: str,
        wasm_module_name: str,
        wasm_module_handler: WasmModuleHandler,
    ) -> Generator[WasmModuleRef, None, None]:
        with create_project(project_name) as proj_ref:
            wasm_ref = qnx.wasm_modules.upload(
                wasm_module_handler=wasm_module_handler,
                name=wasm_module_name,
                project=proj_ref,
            )

            yield wasm_ref

    return upload_wasm


@pytest.fixture(name="create_gpu_decoder_config_in_project")
def fixture_create_gpu_decoder_config_in_project(
    create_project: Callable[[str], ContextManager[ProjectRef]],
) -> Callable[[str, str, str], ContextManager[GpuDecoderConfigRef]]:
    """Returns a `ContextManager` that yields a `GpuDecoderConfigRef` to the `gpu_decoder_config`
    uploaded to a project with `project_name`.

    Notes: will create the project if it does not exist."""

    @contextmanager
    def upload_gpu_decoder_config(
        project_name: str,
        gpu_decoder_config_name: str,
        gpu_decoder_config: str,
    ) -> Generator[GpuDecoderConfigRef, None, None]:
        with create_project(project_name) as proj_ref:
            gpu_decoder_config_ref = qnx.gpu_decoder_configs.upload(
                gpu_decoder_config=gpu_decoder_config,
                name=gpu_decoder_config_name,
                project=proj_ref,
            )

            yield gpu_decoder_config_ref

    return upload_gpu_decoder_config


@pytest.fixture(name="test_ref_serialisation")
def fixture_test_ref_serialisation(  # type: ignore[no-untyped-def]
    tmpdir,
) -> Callable[[str, Ref], None]:
    """Returns a function that tests the serialisation (save/load) of
    a reference."""

    def test_ref_serialisation(
        ref_type: str,
        ref: Ref,
    ) -> None:
        ref_path = tmpdir / ref_type
        save(ref=ref, path=ref_path)
        ref_loaded = load(path=ref_path)
        assert ref == ref_loaded

    return test_ref_serialisation


@pytest.fixture(name="qa_hugr_package")
def qa_hugr_package_fixture() -> Package:
    hugr_path = Path("tests/data/example.hugr").resolve()
    return Package.from_bytes(hugr_path.read_bytes())


@pytest.fixture(name="qa_wasm_module")
def qa_wasm_module_fixture() -> WasmFileHandler:
    wasm_path = Path("examples/basics/data/add_one.wasm").resolve()
    return WasmFileHandler(filepath=str(wasm_path))


@pytest.fixture(name="qa_qir_bitcode")
def qa_qir_bitcode_fixture(test_circuit: Circuit, test_case_name: str) -> bytes:
    qir_bitcode = pytket_to_qir(test_circuit, name=f"qir for {test_case_name}")
    assert isinstance(qir_bitcode, bytes)
    return qir_bitcode


@pytest.fixture(name="qa_gpu_decoder_config")
def qa_gpu_decoder_config_fixture() -> str:
    gpu_decoder_config_path = Path(
        "examples/basics/data/gpu_decoder_config.yaml"
    ).resolve()
    with open(gpu_decoder_config_path) as fp:
        return fp.read()


def get_backend_config_name(backend_config: qnx.BackendConfig) -> str:
    name = backend_config.__class__.__name__

    if hasattr(backend_config, "backend_name"):
        name += f"({backend_config.backend_name})"
    if hasattr(backend_config, "device_name"):
        name += f"({backend_config.device_name})"
    return name


@pytest.fixture(
    scope="session",
    params=[
        # Nexus-hosted
        qnx.AerConfig(),
        qnx.AerStateConfig(),
        qnx.AerUnitaryConfig(),
        qnx.BraketConfig(local=True),
        qnx.QuantinuumConfig(device_name="H1-1LE"),
        qnx.QulacsConfig(),
        # Non Nexus-hosted
        qnx.IBMQConfig(
            backend_name="ibm_brisbane",
            instance=(
                "crn:v1:bluemix:public:quantum-computing:us-east:"
                "a/18f63f4565ef4a40851959792418cbf2:"
                "37bf946a-6349-47df-a092-3fbd5b92dbf2::"
            ),
        ),
        qnx.IBMQEmulatorConfig(
            backend_name="ibm_brisbane",
            instance=(
                "crn:v1:bluemix:public:quantum-computing:us-east:"
                "a/18f63f4565ef4a40851959792418cbf2:"
                "37bf946a-6349-47df-a092-3fbd5b92dbf2::"
            ),
        ),
        qnx.QuantinuumConfig(device_name="H1-1SC"),  # Cluster-hosted
    ],
    ids=get_backend_config_name,
)
def backend_config(request: pytest.FixtureRequest) -> BackendConfig:
    """Fixture to provide an instance of all BackendConfigs for testing."""
    return cast(BackendConfig, request.param)


@pytest.fixture()
def test_circuit() -> Circuit:
    """A pytket circuit to use in the integration tests."""

    circuit = Circuit(3)
    circuit.H(0)
    for i, j in zip(circuit.qubits[:-1], circuit.qubits[1:]):
        circuit.CX(i, j)
    circuit.measure_all()
    return circuit


def _get_or_create_circuit(
    circuit: Circuit,
    project_name: str,
    circuit_name: str,
) -> CircuitRef:
    """Get or create a circuit with `circuit_name` in a project with
    `project_name`."""

    project_ref = qnx.projects.get_or_create(project_name)

    # Using `qnx.circuits.get_all()` since the names of compiled
    # circuits are the same as the original circuits with some suffix,
    # which makes `qnx.circuits.get(namelike)` to return multiple circuits
    circuits_df = qnx.circuits.get_all(project=project_ref).df()

    circuits_filtered = pd.DataFrame()
    if len(circuits_df):
        circuits_filtered = circuits_df.loc[circuits_df["name"] == circuit_name]

    if len(circuits_filtered):
        circuit_ref = qnx.circuits.get(id=str(circuits_filtered.id.values[0]))
    else:
        circuit_ref = qnx.circuits.upload(
            circuit=circuit,
            name=circuit_name,
            description=f"description for {circuit_name}",
            project=project_ref,
        )

    return circuit_ref



"""Test basic functionality relating to the user module."""

from typing import Callable

import qnexus as qnx
from qnexus.models.references import Ref, UserRef


def test_user_get(authenticated_nexus: None) -> None:
    """Test that we can get a specific unique UserRef."""

    my_user = qnx.users.get_self()
    assert isinstance(my_user, UserRef)

    my_user_again = qnx.users._fetch_by_id(user_id=my_user.id)
    assert isinstance(my_user_again, UserRef)
    assert my_user == my_user_again


def test_user_ref_serialisation(
    authenticated_nexus: None,
    test_ref_serialisation: Callable[[str, Ref], None],
) -> None:
    """Test the serialisation round trip of a UserRef."""

    my_user = qnx.users.get_self()

    test_ref_serialisation("user", my_user)



"""The qnexus package."""

from __future__ import annotations

from collections import OrderedDict
from datetime import datetime
from typing import Any, TypedDict

import pandas as pd
from pydantic import BaseModel, ConfigDict, Field, field_serializer, field_validator

PropertiesDict = OrderedDict[str, bool | int | float | str]


class AnnotationsDict(TypedDict, total=False):
    """TypedDict for annotations"""

    name: str | None
    description: str | None
    properties: PropertiesDict
    created: datetime | None
    modified: datetime | None


class Annotations(BaseModel):
    """Pydantic model for annotations"""

    name: str | None = None
    description: str | None = None
    properties: PropertiesDict = Field(default_factory=OrderedDict)
    created: datetime | None = None
    modified: datetime | None = None

    model_config = ConfigDict(frozen=True)

    @field_validator("properties")
    @classmethod
    def sort_properties(cls, v: PropertiesDict) -> PropertiesDict:
        """Sort the values of the properties dict."""
        return PropertiesDict(OrderedDict(sorted(v.items())))

    @field_serializer("created")
    def serialize_created(self, created: datetime | None) -> str | None:
        """Custom serializer for datetimes."""
        if created:
            return str(created)
        return None

    @field_serializer("modified")
    def serialize_modified(self, modified: datetime | None) -> str | None:
        """Custom serializer for datetimes."""
        if modified:
            return str(modified)
        return None

    def df(self) -> pd.DataFrame:
        """Convert to a pandas DataFrame."""
        return pd.DataFrame(
            {
                "name": self.name,
                "description": self.description,
                "created": self.created,
                "modified": self.modified,
            }
            | self.properties,
            index=[0],
        )

    @classmethod
    def from_dict(cls, annotations_dict: dict[str, Any]) -> Annotations:
        """Construct Annotations from a dict."""
        return Annotations(
            name=annotations_dict["name"],
            description=annotations_dict.get("description", None),
            properties=PropertiesDict(**annotations_dict.get("properties", {})),
            created=annotations_dict["timestamps"]["created"],
            modified=annotations_dict["timestamps"]["modified"],
        )


class CreateAnnotations(BaseModel):
    """Pydantic model for annotations when the name is required."""

    name: str
    description: str | None = None
    properties: PropertiesDict | None = Field(default_factory=PropertiesDict)

    model_config = ConfigDict(frozen=True)

    @field_validator("properties")
    @classmethod
    def set_properties_default(cls, v: PropertiesDict | None) -> PropertiesDict:
        """Replace None properties with an empty PropertiesDict on model construction."""
        if v is None:
            return PropertiesDict()
        return v



"""Utilities for qnexus-filesystem interactions."""

import json
from pathlib import Path

from qnexus.models.references import Ref, deserialize_nexus_ref


def save(ref: Ref, path: Path, mkdir: bool = False) -> None:
    """Save a Nexus Ref to a file."""
    if mkdir:
        path.parent.mkdir(parents=True, exist_ok=True)
    with open(path, "w", encoding="utf-8") as f:
        f.write(ref.model_dump_json())


def load(path: Path) -> Ref:
    """Load a Nexus Ref from a file."""
    with open(path, "r", encoding="utf-8") as f:
        data = json.load(f)

    return deserialize_nexus_ref(data)



"""Basic checks for API query filters."""

from datetime import datetime
from uuid import uuid4

from qnexus.models.annotations import Annotations, PropertiesDict
from qnexus.models.filters import (
    ArchivedFilter,
    CreatorFilter,
    CredentialsFilter,
    DevicesFilter,
    FuzzyNameFilter,
    JobStatusFilter,
    JobTypeFilter,
    PaginationFilter,
    ProjectRefFilter,
    PropertiesFilter,
    ScopeFilter,
    ScopeFilterEnum,
    SortFilter,
    SortFilterEnum,
    TimeFilter,
)
from qnexus.models.job_status import JobStatusEnum
from qnexus.models.references import JobType, ProjectRef


def test_all_filter_serialisation() -> None:
    """Test all supported filters and their serialisation."""

    class Params(
        ArchivedFilter,
        CreatorFilter,
        CredentialsFilter,
        DevicesFilter,
        FuzzyNameFilter,
        JobStatusFilter,
        JobTypeFilter,
        PaginationFilter,
        ProjectRefFilter,
        PropertiesFilter,
        ScopeFilter,
        SortFilter,
        TimeFilter,
    ):
        """Test filters class."""

    dummy_project_ref = ProjectRef(
        id=uuid4(), annotations=Annotations(), contents_modified=datetime.now()
    )

    # Mix between strings and enums, both should be supported
    job_status = [JobStatusEnum.COMPLETED, "CANCELLED"]
    job_type = [JobType.EXECUTE, "compile"]
    sort_filters = [SortFilterEnum.CREATED_DESC, "-modified"]
    test_datetime = datetime.now()

    params = Params(
        name_like="test_name",
        creator_email=["test@email.com"],
        project=dummy_project_ref,
        status=(
            JobStatusFilter.convert_status_filters(job_status)  # type: ignore
            if job_status
            else None
        ),
        job_type=job_type,
        properties=PropertiesDict({"hello": 1, "goodbye": False, "how": "yes"}),
        created_before=test_datetime,
        created_after=test_datetime,
        modified_before=test_datetime,
        modified_after=test_datetime,
        sort=SortFilter.convert_sort_filters(sort_filters),  # type: ignore
        page_number=1,
        page_size=100,
        scope=ScopeFilterEnum.ORG_ADMIN,
    ).model_dump(by_alias=True, exclude_unset=True, exclude_none=True)

    assert params["filter[name]"] == "test_name"
    assert params["filter[creator][email]"] == ["test@email.com"]
    assert params["filter[project][id]"] == str(dummy_project_ref.id)
    assert sorted(params["filter[properties]"]) == sorted(
        ["(hello,1)", "(goodbye,false)", '(how,"yes")']
    )
    assert sorted(params["filter[status][status]"]) == sorted(
        ["COMPLETED", "CANCELLED"]
    )
    assert sorted(params["filter[job_type]"]) == sorted(["execute", "compile"])
    assert params["filter[timestamps][created][before]"] == test_datetime
    assert params["filter[timestamps][created][after]"] == test_datetime
    assert params["filter[timestamps][modified][before]"] == test_datetime
    assert params["filter[timestamps][modified][after]"] == test_datetime
    assert sorted(params["sort"]) == sorted(
        ["-timestamps.created", "-timestamps.modified"]
    )
    assert params["page[number]"] == 1
    assert params["page[size]"] == 100
    assert params["scope"] == "org_admin"



"""Utlity functions for the client."""

import http
import json
import os
import warnings
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Literal, ParamSpec, TypeVar

from httpx import Response
from pydantic import BaseModel

import qnexus.exceptions as qnx_exc
from qnexus.config import CONFIG

TokenTypes = Literal["access_token", "refresh_token"]

token_file_from_type = {
    "access_token": "id.json",
    "refresh_token": "token.json",
}


def normalize_included(included: list[Any]) -> dict[str, dict[str, Any]]:
    """Convert a JSON API included array into a mapped dict of the form:
    {
        "user": {
            [user_id]: User
        },
        "project": {
            [project_id]: Project
        }
    }
    """
    included_map: dict[str, dict[str, Any]] = {}
    for item in included:
        included_map.setdefault(item["type"], {item["id"]: {}})
        included_map[item["type"]][item["id"]] = item
    return included_map


def remove_token(token_type: TokenTypes) -> None:
    """Delete a token file."""
    # Don't try to delete refresh token in Jupyterhub
    if is_jupyterhub_environment() and token_type == "refresh_token":
        return
    token_file_path = Path.home() / CONFIG.token_path / token_file_from_type[token_type]
    if token_file_path.exists():
        token_file_path.unlink()


class RefreshTokenData(BaseModel):
    """Stored refresh token data."""

    delete_version_after: str | None
    refresh_token: str


class RefreshToken(BaseModel):
    """Model for token storage file."""

    data: RefreshTokenData


class AccessTokenData(BaseModel):
    """Stored access token data."""

    access_token: str


class AccessToken(BaseModel):
    """Model for access token storage."""

    data: AccessTokenData


def read_token(token_type: TokenTypes) -> str:
    """Read a token from a file."""
    token_file_path = Path.home() / CONFIG.token_path
    with (token_file_path / token_file_from_type[token_type]).open(
        encoding="UTF-8"
    ) as file:
        file_contents = file.read().strip()
        if token_type == "access_token":
            return AccessToken(**json.loads(file_contents)).data.access_token
        return RefreshToken(**json.loads(file_contents)).data.refresh_token


def write_token(token_type: TokenTypes, token: str) -> None:
    """Write a token to a file."""

    # don't allow writing of refresh token in Jupyterhub
    if is_jupyterhub_environment() and token_type == "refresh_token":
        return

    token_file_path = Path.home() / CONFIG.token_path
    token_file_path.mkdir(parents=True, exist_ok=True)
    with (token_file_path / token_file_from_type[token_type]).open(
        encoding="UTF-8", mode="w"
    ) as file:
        if token_type == "access_token":
            file.write(
                AccessToken(data=AccessTokenData(access_token=token)).model_dump_json()
            )
            return
        file.write(
            RefreshToken(
                data=RefreshTokenData(refresh_token=token, delete_version_after=None)
            ).model_dump_json()
        )


def consolidate_error(res: Response, description: str) -> None:
    """Consolidate as much error-checking of response"""
    # check if token has expired or is generally unauthorized
    resp_json = res.json()
    if res.status_code == http.HTTPStatus.UNAUTHORIZED:
        raise qnx_exc.AuthenticationError(
            (
                f"Authorization failure attempting: {description}."
                f"\n\nServer Response: {resp_json}"
            )
        )
    if res.status_code != http.HTTPStatus.OK:
        raise qnx_exc.AuthenticationError(
            f"HTTP error attempting: {description}.\n\nServer Response: {resp_json}"
        )


def handle_fetch_errors(res: Response) -> None:
    """Handle errors related to a fetch request."""

    if res.status_code == 404:
        raise qnx_exc.ZeroMatches()

    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(message=res.text, status_code=res.status_code)


def is_jupyterhub_environment() -> bool:
    """Check if the module is running in the Nexus JupyterHub."""
    if os.environ.get("JUPYTERHUB_USER"):
        return True
    return False


ParamsWithProgramsOrCircuits = ParamSpec("ParamsWithProgramsOrCircuits")
ReturnType = TypeVar("ReturnType")


def accept_circuits_for_programs(
    fn: Callable[ParamsWithProgramsOrCircuits, ReturnType],
) -> Callable[ParamsWithProgramsOrCircuits, ReturnType]:
    """
    Nexus's /api/jobs/v1beta3 introduces the concept of programs for compilation
    and execution. This decorator shims from an argument called "circuits" to an
    argument called "programs", emitting a deprecation warning if necessary, so
    that the pre-existing qnexus python API can support both methods for a
    while.
    """

    @wraps(fn)
    def wrapper(
        *args: ParamsWithProgramsOrCircuits.args,
        **kwargs: ParamsWithProgramsOrCircuits.kwargs,
    ) -> ReturnType:
        if "circuits" in kwargs:
            kwargs["programs"] = kwargs["circuits"]
            warnings.warn(
                "The `circuits` argument is deprecated and will be removed in a "
                "future version. Please use `programs`.",
                category=DeprecationWarning,
            )
            del kwargs["circuits"]
        return fn(*args, **kwargs)

    return wrapper



import datetime as dt
import warnings
from typing import Union
from unittest import mock
from uuid import uuid4

from qnexus import QuantinuumConfig
from qnexus.client.jobs._compile import start_compile_job
from qnexus.client.utils import accept_circuits_for_programs
from qnexus.models.references import CircuitRef, ProjectRef

PROJECT_REF = ProjectRef(
    annotations={}, id=uuid4(), contents_modified=dt.datetime.now()
)
CIRCUIT_REF = CircuitRef(
    id=uuid4(),
    annotations={},
    project=PROJECT_REF,
)


def test_wrapper() -> None:
    """
    Test that `accept_circuits_for_programs` decorator invokes the wrapped
    function as expected, emitting a deprecation warning when the old `circuits`
    keyword arg is used.
    """

    @accept_circuits_for_programs
    def inner_fn(
        name: str,
        programs: Union[CircuitRef, list[CircuitRef]],
    ) -> None:
        assert programs == [CIRCUIT_REF]

    with warnings.catch_warnings(record=True) as captured:
        inner_fn("hello", [CIRCUIT_REF])
        assert len(captured) == 0

    with warnings.catch_warnings(record=True) as captured:
        inner_fn("hello", circuits=[CIRCUIT_REF])  # type: ignore
        assert len(captured) == 1
        assert captured[0].category is DeprecationWarning

    with warnings.catch_warnings(record=True) as captured:
        inner_fn("hello", programs=[CIRCUIT_REF])
        assert len(captured) == 0


def test_compile_circuit_with_wrapper() -> None:
    """
    Test that `start_compile_job` is correctly invoked when the deprecated
    `circuits` keyword argument is supplied. The `accept_circuits_for_programs`
    decorator implements this.
    """
    with mock.patch("qnexus.client.jobs._compile.get_nexus_client") as gnc:
        mock_client = mock.MagicMock()

        mock_resp = mock.MagicMock()
        mock_resp.status_code = 202
        mock_resp.json.return_value = {
            "data": {
                "id": str(uuid4()),
                "attributes": {
                    "name": "blah",
                    "timestamps": {
                        "created": dt.datetime.now(),
                        "modified": dt.datetime.now(),
                    },
                },
            }
        }

        mock_client.post.return_value = mock_resp
        gnc.return_value = mock_client

        gnc.post.return_value = mock_resp

        with warnings.catch_warnings(record=True) as captured:
            start_compile_job(
                name="foo",
                backend_config=QuantinuumConfig(device_name="whatever"),
                circuits=[CIRCUIT_REF],  # type: ignore
                project=PROJECT_REF,
            )
            assert captured[0].category is DeprecationWarning

        assert mock_client.post.call_count == 1
        assert (
            "program_id"
            in mock_client.post.call_args[1]["json"]["data"]["attributes"][
                "definition"
            ]["items"][0]
        )



"""Quantinuum Nexus API client configuration via pydantic-settings."""

from pydantic import AliasChoices, Field
from pydantic_settings import BaseSettings, SettingsConfigDict


class Config(BaseSettings):
    """QNexus Configuration schema.

    Uses pydantic-settings to read environment variables for qnexus configuration."""

    model_config = SettingsConfigDict(env_prefix="NEXUS_")

    # web
    protocol: str = "https"
    websockets_protocol: str = "wss"
    domain: str = Field(
        validation_alias=AliasChoices("NEXUS_DOMAIN", "NEXUS_HOST"),
        default="nexus.quantinuum.com",
    )
    port: int = 443
    httpx_verify: bool = True

    # auth
    store_tokens: bool = True  # Not implemented
    token_path: str = ".qnx/auth"

    # testing
    qa_user_email: str = ""
    qa_user_password: str = ""

    def __str__(self) -> str:
        """String representation of current config."""
        return self.model_dump_json()

    @property
    def url(self) -> str:
        """Current http API URL"""
        return f"{self.protocol}://{self.domain}:{self.port}"

    @property
    def websockets_url(self) -> str:
        """Current websockets API URL"""
        return f"{self.websockets_protocol}://{self.domain}:{self.port}"


CONFIG = Config()



"""Client API for quotas in Nexus."""

from typing import Literal

from qnexus.client import get_nexus_client
from qnexus.client.users import get_self
from qnexus.exceptions import ResourceFetchFailed
from qnexus.models import Quota
from qnexus.models.references import DataframableList

QuotaName = Literal["compilation", "simulation", "jupyterhub", "database_usage"]

_quota_map = {
    "compilation": "total_time_taken",
    "simulation": "total_time_taken",
    "jupyterhub": "total_time_taken",
    "database_usage": "megabytes_used",
}

NO_QUOTA_SET = "No quota set for user"


def get_all() -> DataframableList[Quota]:
    """Get all quotas, including usage."""
    user = get_self()

    res = get_nexus_client().get(
        "/api/quotas/v1beta",
        params={
            "entity_type": "user",
            "entity_id": str(user.id),
            "include_usage": True,
        },
    )

    if res.status_code != 200:
        raise ResourceFetchFailed(message=res.text, status_code=res.status_code)

    quota_list: DataframableList[Quota] = DataframableList([])
    for quota in res.json():
        try:
            quota_key = _quota_map[quota["quota"]["name"]]
        except KeyError:
            # If the quota name is not in the map, skip it
            continue
        quota_value = quota["quota"]["details"].get(quota_key, None)
        quota_list.append(
            Quota(
                name=quota["quota"]["name"],
                description=quota["quota"]["details"]["description"],
                usage=quota["quota"]["usage"].get(quota_key, 0),
                quota=quota_value if quota_value else NO_QUOTA_SET,
            )
        )

    return quota_list


def get(name: QuotaName) -> Quota:
    """Get specific quota details by name."""
    user = get_self()

    res = get_nexus_client().get(
        "/api/quotas/v1beta3",
        params={
            "entity_type": "user",
            "entity_id": str(user.id),
            "name": name,
            "include_usage": True,
        },  # needs user id
    )

    if res.status_code != 200:
        raise ResourceFetchFailed(message=res.text, status_code=res.status_code)

    quota = res.json()[0]

    quota_key = _quota_map[quota["quota"]["name"]]
    quota_value = quota["quota"]["details"].get(quota_key, None)

    return Quota(
        name=quota["quota"]["name"],
        description=quota["quota"]["details"]["description"],
        usage=quota["quota"]["usage"].get(quota_key, None),
        quota=quota_value if quota_value else NO_QUOTA_SET,
    )


def check_quota(name: QuotaName) -> bool:
    """Check that the current user has available quota."""
    user = get_self()
    res = get_nexus_client().get(
        "/api/quotas/v1beta3/guard", params={"name": name, "user_id": str(user.id)}
    )

    if res.status_code != 200:
        return False
    return True



"""Custom exceptions thrown in qnexus usage."""

from typing import Optional


class AuthenticationError(Exception):
    "Raised when there is an issue authenticating with the Nexus API."


class ResourceFetchFailed(Exception):
    """
    Raised when a resource cannot be fetched from the platform to be read, or
    when expected data wasn't found on an object.
    """

    def __init__(self, message: str, status_code: Optional[int] = None) -> None:
        self.message = message
        self.status_code = status_code
        self.err = (
            "Failed to fetch resource with status code: "
            f"{self.status_code}, message: {self.message}"
        )
        super().__init__(self.err)


class ZeroMatches(Exception):
    """Raised when a ``get`` call finds no matches."""


class NoUniqueMatch(Exception):
    """Raised when a ``get`` call finds more than one match."""


class JobError(Exception):
    """Raised when a Nexus Job has errored."""


class ResourceCreateFailed(Exception):
    """Raised when a resource couldn't be created on the platform."""

    def __init__(self, message: str, status_code: Optional[int] = None) -> None:
        self.message = message
        self.status_code = status_code
        self.err = (
            "Failed to create resource with status code: "
            f"{self.status_code}, message: {self.message}"
        )
        super().__init__(self.err)


class ResourceDeleteFailed(Exception):
    """Raised when a resource couldn't be deleted on the platform."""

    def __init__(self, message: str, status_code: Optional[int] = None) -> None:
        self.message = message
        self.status_code = status_code
        self.err = (
            "Failed to delete resource with status code: "
            f"{self.status_code}, message: {self.message}"
        )
        super().__init__(self.err)


class ResourceUpdateFailed(Exception):
    """Raised when a resource cannot be updated on the platform."""

    def __init__(self, message: str, status_code: Optional[int] = None) -> None:
        self.message = message
        self.status_code = status_code
        self.err = (
            "Failed to update resource with status code: "
            f"{self.status_code}, message: {self.message}"
        )
        super().__init__(self.err)


class IncompatibleResultVersion(Exception):
    """Raised when the user tries to get results in a version that's not possible"""



"""Test basic functionality relating to the qir module."""

from collections import Counter
from pathlib import Path
from typing import Callable, ContextManager

import pyqir
from hugr.qsystem.result import QsysResult
from pytket.backends.backendinfo import BackendInfo
from pytket.backends.backendresult import BackendResult
from pytket.circuit import Bit, Circuit
from pytket.qir import pytket_to_qir  # type: ignore[attr-defined]

import qnexus as qnx
from qnexus.models.annotations import PropertiesDict
from qnexus.models.references import ProjectRef, QIRRef, QIRResult, Ref, ResultVersions


def test_qir_create_and_update(
    test_case_name: str,
    create_property_in_project: Callable[..., ContextManager[ProjectRef]],
    test_circuit: Circuit,
) -> None:
    """Test that we can create a qir and add a property value."""

    project_name = f"project for {test_case_name}"
    property_name = f"property for {test_case_name}"

    with create_property_in_project(
        project_name=project_name,
        property_name=property_name,
        property_type="string",
        required=False,
    ):
        proj_ref = qnx.projects.get_or_create(project_name)
        qir_name = f"qir for {test_case_name}"

        qir_bitcode = pytket_to_qir(test_circuit, name=qir_name)
        assert isinstance(qir_bitcode, bytes)
        my_new_qir = qnx.qir.upload(qir=qir_bitcode, name=qir_name, project=proj_ref)

        assert isinstance(my_new_qir, QIRRef)

        test_prop_value = "foo"

        updated_qir_ref = qnx.qir.update(
            ref=my_new_qir,
            properties=PropertiesDict({property_name: test_prop_value}),
        )

        assert updated_qir_ref.annotations.properties[property_name] == test_prop_value


def test_qir_download(
    test_case_name: str,
    create_qir_in_project: Callable[[str, str, bytes], ContextManager[QIRRef]],
    qa_qir_bitcode: bytes,
) -> None:
    """Test that QIR bytes can be downloaded from an uploaded QIR module."""

    project_name = f"project for {test_case_name}"
    qir_name = f"qir for {test_case_name}"

    with create_qir_in_project(
        project_name,
        qir_name,
        qa_qir_bitcode,
    ) as qir_ref:
        qir_bytes = qir_ref.download_qir()
        assert isinstance(qir_bytes, bytes)
        assert qir_bytes == qa_qir_bitcode


def test_qir_get_by_id(
    test_case_name: str,
    create_qir_in_project: Callable[[str, str, bytes], ContextManager[QIRRef]],
    qa_qir_bitcode: bytes,
    test_ref_serialisation: Callable[[str, Ref], None],
) -> None:
    """Test that we can get a QIRRef by its ID and its round trip serialisation."""

    project_name = f"project for {test_case_name}"
    qir_name = f"qir for {test_case_name}"

    with create_qir_in_project(
        project_name,
        qir_name,
        qa_qir_bitcode,
    ):
        my_proj = qnx.projects.get(name_like=project_name)
        my_qir_ref = qnx.qir.get(name_like=qir_name, project=my_proj)

        qir_ref_by_id = qnx.qir.get(id=my_qir_ref.id)

        assert qir_ref_by_id == my_qir_ref

        test_ref_serialisation("qir", qir_ref_by_id)


def test_qir_get_all(
    test_case_name: str,
    create_qir_in_project: Callable[[str, str, bytes], ContextManager[QIRRef]],
    qa_qir_bitcode: bytes,
) -> None:
    """Test that we can get all qirRefs in a project."""

    project_name = f"project for {test_case_name}"
    qir_name = f"qir for {test_case_name}"

    with create_qir_in_project(
        project_name,
        qir_name,
        qa_qir_bitcode,
    ):
        my_proj = qnx.projects.get(name_like=project_name)

        qirs = qnx.qir.get_all(project=my_proj)

        assert qirs.count() >= 1
        assert isinstance(qirs.list()[0], QIRRef)


def test_execution(
    test_case_name: str,
    create_qir_in_project: Callable[[str, str, bytes], ContextManager[QIRRef]],
    qa_qir_bitcode: bytes,
) -> None:
    """Test the execution of a QIR program."""

    project_name = f"project for {test_case_name}"
    qir_name = f"qir for {test_case_name}"

    with create_qir_in_project(
        project_name,
        qir_name,
        qa_qir_bitcode,
    ) as qir_program_ref:
        device_name = "H1-1SC"  # Syntax checker - no results

        project_ref = qnx.projects.get_or_create(name=project_name)

        qir_program_ref = qnx.qir.get(name_like=qir_name)

        job_ref = qnx.start_execute_job(
            programs=[qir_program_ref],
            n_shots=[10],
            backend_config=qnx.QuantinuumConfig(device_name=device_name),
            project=project_ref,
            name=f"qir job for {test_case_name}",
        )

        qnx.jobs.wait_for(job_ref)

        results = qnx.jobs.results(job_ref)

        assert len(results) == 1
        result_ref = results[0]

        assert isinstance(result_ref.download_backend_info(), BackendInfo)
        assert isinstance(result_ref.get_input(), QIRRef)

        assert result_ref.get_input().id == qir_program_ref.id

        qir_result = qnx.jobs.results(job_ref)[0].download_result()
        assert isinstance(qir_result, BackendResult)
        assert qir_result.get_counts() == Counter({(0, 0, 0): 10})
        assert qir_result.get_bitlist() == [Bit("c", 2), Bit("c", 1), Bit("c", 0)]


def test_execution_on_NG_devices(
    test_case_name: str,
    create_qir_in_project: Callable[[str, str, bytes], ContextManager[QIRRef]],
) -> None:
    """Test execution on NG devices, specifically to focus on getting the results"""

    project_name = f"project for {test_case_name}"
    qir_name = f"qir for {test_case_name}"

    with create_qir_in_project(
        project_name,
        qir_name,
        make_qir_bitcode_from_file("base.ll"),
    ) as qir_ref:
        project_ref = qnx.projects.get(name_like=project_name)

        job_ref = qnx.start_execute_job(
            programs=[qir_ref],
            n_shots=[10],
            backend_config=qnx.QuantinuumConfig(device_name="Helios-1E", max_cost=10),
            project=project_ref,
            name=f"qir job for {test_case_name}",
        )

        qnx.jobs.wait_for(job_ref)

        results = qnx.jobs.results(job_ref)[0].download_result()
        # Assert this is a QIR compliant result
        assert isinstance(results, QIRResult)
        escaped_results = results.results.encode("unicode_escape").decode()
        assert "HEADER\\tschema_id\\tlabeled" in escaped_results
        # Can't assert the value is the same, so just check the output is there
        assert "OUTPUT\\tTUPLE\\t2\\tt0" in escaped_results

        v4_results = qnx.jobs.results(job_ref)[0].download_result(
            version=ResultVersions.RAW
        )
        # Assert this is in v4 format
        assert isinstance(v4_results, QsysResult)
        assert v4_results.results[0].entries[0][0] == "USER:QIRTUPLE:t0"


def make_qir_bitcode_from_file(filename: str) -> bytes:
    with open(
        Path("tests/data").resolve() / filename,
        "r",
    ) as file:
        return pyqir.Module.from_ir(pyqir.Context(), file.read()).bitcode



"""Filter models for use by the client."""

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Annotated, Literal, OrderedDict, Union

from pydantic import BaseModel, ConfigDict, Field, field_serializer

from qnexus.models import CredentialIssuer, JobStatusEnum
from qnexus.models.annotations import PropertiesDict
from qnexus.models.references import JobType, ProjectRef
from qnexus.models.utils import AllowNone


def _format_property(key: str, value: bool | int | float | str) -> str:
    if isinstance(value, str):
        return f'({key},"{value}")'
    if isinstance(value, bool):
        return f"({key},{str(value).lower()})"
    return f"({key},{value})"


class PropertiesFilter(BaseModel):
    """Properties filters model."""

    properties: PropertiesDict | None = Field(
        default=OrderedDict(),
        serialization_alias="filter[properties]",
        description="Filter by resource label value.",
    )

    @field_serializer("properties")
    def serialize_properties(self, properties: PropertiesDict) -> list[str]:
        """Serialize the id for a ProjectRef."""
        return [_format_property(key, value) for key, value in properties.items()]


class TimeFilter(BaseModel):
    """Resource time filters model."""

    created_before: Annotated[datetime | None, AllowNone] = Field(
        default=None,
        serialization_alias="filter[timestamps][created][before]",
        description="Show items created before this date.",
    )
    created_after: Annotated[datetime | None, AllowNone] = Field(
        default=None,
        serialization_alias="filter[timestamps][created][after]",
        description="Show items created after this date.",
    )
    modified_before: Annotated[datetime | None, AllowNone] = Field(
        default=None,
        serialization_alias="filter[timestamps][modified][before]",
        description="Show items modified before this date.",
    )
    modified_after: Annotated[datetime | None, AllowNone] = Field(
        default=None,
        serialization_alias="filter[timestamps][modified][after]",
        description="Show items modified after this date.",
    )


class PaginationFilter(BaseModel):
    """Pagination model."""

    page_number: int | None = Field(
        default=0,
        serialization_alias="page[number]",
        description="Specific page to return.",
    )
    page_size: int | None = Field(
        default=50,
        serialization_alias="page[size]",
        description="Size of page that is returned.",
    )


class CreatorFilter(BaseModel):
    """Creator email model."""

    creator_email: list[str] | None = Field(
        default=[],
        serialization_alias="filter[creator][email]",
        examples=["user@domain.com"],
        description="Filter by creator email.",
    )


class FuzzyNameFilter(BaseModel):
    """Name model."""

    name_like: str | None = Field(
        default="",
        serialization_alias="filter[name]",
        description="Filter by name, fuzzy search.",
    )


class SortFilterEnum(str, Enum):
    """SortFilterEnum model."""

    CREATED_ASC = "created"
    CREATED_DESC = "-created"
    MODIFIED_ASC = "modified"
    MODIFIED_DESC = "-modified"
    NAME_ASC = "name"
    NAME_DESC = "-name"


SortFilterString = Union[
    Literal["timestamps.created"],
    Literal["-timestamps.created"],
    Literal["timestamps.modified"],
    Literal["-timestamps.modified"],
    Literal["name"],
    Literal["-name"],
]

sortfilterenum_to_string: dict[SortFilterEnum, SortFilterString] = {
    SortFilterEnum.CREATED_ASC: "timestamps.created",
    SortFilterEnum.CREATED_DESC: "-timestamps.created",
    SortFilterEnum.MODIFIED_ASC: "timestamps.modified",
    SortFilterEnum.MODIFIED_DESC: "-timestamps.modified",
    SortFilterEnum.NAME_ASC: "name",
    SortFilterEnum.NAME_DESC: "-name",
}


class SortFilter(BaseModel):
    """Resource sorting model."""

    sort: (
        list[
            Union[
                Literal["timestamps.created"],
                Literal["-timestamps.created"],
                Literal["timestamps.modified"],
                Literal["-timestamps.modified"],
                Literal["name"],
                Literal["-name"],
            ]
        ]
        | None
    ) = Field(
        default=["-timestamps.created"],  # type: ignore
        serialization_alias="sort",
        description="Sort items.",
    )

    @staticmethod
    def convert_sort_filters(
        sort_filters: list[SortFilterEnum] | None,
    ) -> list[SortFilterString] | None:
        """Convert SortFilterEnum to SortFilterString."""
        return (
            [sortfilterenum_to_string[sort_filter] for sort_filter in sort_filters]
            if sort_filters
            else None
        )


class ProjectRefFilter(BaseModel):
    """Project Id filter"""

    project: Annotated[ProjectRef | None, AllowNone] = Field(
        default=None,
        serialization_alias="filter[project][id]",
        description="Filter by project ref",
    )

    @field_serializer("project")
    def serialize_project_ref(self, project: ProjectRef) -> str:
        """Serialize the id for a ProjectRef."""
        return str(project.id)


class ArchivedFilter(BaseModel):
    """Include or omit archived projects"""

    is_archived: Annotated[bool, AllowNone] = Field(
        default=False,
        serialization_alias="filter[archived]",
        description="Include or omit archived projects",
    )


JobStatusString = Union[
    Literal["COMPLETED"],
    Literal["QUEUED"],
    Literal["SUBMITTED"],
    Literal["RUNNING"],
    Literal["CANCELLED"],
    Literal["ERROR"],
    Literal["DEPLETED"],
    Literal["TERMINATED"],
]

jobstatusenum_to_string: dict[JobStatusEnum, JobStatusString] = {
    JobStatusEnum.COMPLETED: "COMPLETED",
    JobStatusEnum.QUEUED: "QUEUED",
    JobStatusEnum.SUBMITTED: "SUBMITTED",
    JobStatusEnum.RUNNING: "RUNNING",
    JobStatusEnum.CANCELLED: "CANCELLED",
    JobStatusEnum.ERROR: "ERROR",
    JobStatusEnum.DEPLETED: "DEPLETED",
    JobStatusEnum.TERMINATED: "TERMINATED",
}


class JobStatusFilter(BaseModel):
    """Job status filter"""

    status: list[JobStatusString] | None = Field(
        default=[  # type: ignore
            "COMPLETED",
            "QUEUED",
            "SUBMITTED",
            "RUNNING",
            "CANCELLED",
            "ERROR",
            "DEPLETED",
            "TERMINATED",
        ],
        serialization_alias="filter[status][status]",
        description="Filter by job status",
    )

    @staticmethod
    def convert_status_filters(
        status_filters: list[JobStatusEnum],
    ) -> list[JobStatusString]:
        """Convert SortFilterEnum to SortFilterString."""
        return [
            jobstatusenum_to_string[status_filter] for status_filter in status_filters
        ]


class JobTypeFilter(BaseModel):
    """Filter by job type."""

    job_type: list[JobType] | None = Field(
        default=[JobType.EXECUTE, JobType.COMPILE],
        serialization_alias="filter[job_type]",
        description="Filter by job_type",
    )

    model_config = ConfigDict(use_enum_values=True)


class DevicesFilter(BaseModel):
    """Filter by device details."""

    backend: list[str] | None = None
    region: str | None = None
    ibm_instance: str | None = None
    ibm_region: str | None = None
    credential_ids: list[str] | None = None
    credential_names: list[str] | None = None
    is_local: bool | None = None


class CredentialsFilter(BaseModel):
    """Filter for credentials."""

    issuer: CredentialIssuer | str | None = None


class ScopeFilterEnum(str, Enum):
    """ScopeFilterEnum model."""

    USER = "user"
    ORG_ADMIN = "org_admin"
    GLOBAL_ADMIN = "global_admin"
    HIGHEST = "highest"


class ScopeFilter(BaseModel):
    """Filter by scope."""

    scope: ScopeFilterEnum | None

    model_config = ConfigDict(use_enum_values=True)



"""Client API for users in Nexus."""

from uuid import UUID

import qnexus.exceptions as qnx_exc
from qnexus.client import get_nexus_client
from qnexus.models.references import UserRef


def get_self() -> UserRef:
    """Get the logged in user."""

    res = get_nexus_client().get("/api/users/v1beta2/me")

    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(message=res.text, status_code=res.status_code)

    user_dict = res.json()

    return UserRef(
        display_name=user_dict["data"]["attributes"]["display_name"],
        id=user_dict["data"]["id"],
    )


def _fetch_by_id(user_id: UUID) -> UserRef:
    """Get a specific user."""

    res = get_nexus_client().get(f"/api/users/v1beta/{user_id}")

    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(message=res.text, status_code=res.status_code)

    user_dict = res.json()

    return UserRef(
        display_name=user_dict["data"]["attributes"]["display_name"],
        id=user_dict["data"]["id"],
    )



"""Test basic functionality relating to the wasm_modules module."""

from typing import Callable, ContextManager

import pandas as pd
from pytket.circuit import Circuit
from pytket.wasm.wasm import WasmFileHandler, WasmModuleHandler

import qnexus as qnx
from qnexus.models.job_status import JobStatusEnum
from qnexus.models.references import Ref, WasmModuleRef


def test_wasm_download(
    test_case_name: str,
    create_wasm_in_project: Callable[
        [str, str, WasmModuleHandler], ContextManager[WasmModuleRef]
    ],
    qa_wasm_module: WasmFileHandler,
    test_ref_serialisation: Callable[[str, Ref], None],
) -> None:
    """Test that valid WASM can be extracted from an uploaded WASM module,
    and the WasmModuleRef serialisation round trip."""

    project_name = f"project for {test_case_name}"
    wasm_module_name = f"wasm for {test_case_name}"

    with create_wasm_in_project(
        project_name,
        wasm_module_name,
        qa_wasm_module,
    ) as wasm_ref:
        assert isinstance(wasm_ref, WasmModuleRef)
        downloaded_wasm_module_handler = wasm_ref.download_wasm_contents()

        downloaded_wasm_module_handler.check()
        assert downloaded_wasm_module_handler.functions == qa_wasm_module.functions

        wasm_ref_by_id = qnx.wasm_modules.get(id=wasm_ref.id)
        test_ref_serialisation("wasm", wasm_ref_by_id)


def test_wasm_flow(
    test_case_name: str,
    create_wasm_in_project: Callable[
        [str, str, WasmModuleHandler], ContextManager[WasmModuleRef]
    ],
    qa_wasm_module: WasmFileHandler,
) -> None:
    """Test the flow for executing a simple WASM circuit on H1-1LE."""

    project_name = f"project for {test_case_name}"
    wasm_module_name = f"wasm for {test_case_name}"

    with create_wasm_in_project(
        project_name,
        wasm_module_name,
        qa_wasm_module,
    ) as wasm_ref:
        proj_ref = qnx.projects.get(name_like=project_name)

        my_wasm_db_matches = qnx.wasm_modules.get_all()
        assert isinstance(my_wasm_db_matches.summarize(), pd.DataFrame)
        assert isinstance(next(my_wasm_db_matches), WasmModuleRef)

        wasm_ref = qnx.wasm_modules.get(name_like=wasm_module_name)
        assert isinstance(wasm_ref, WasmModuleRef)

        wasm_ref_2 = qnx.wasm_modules.get(id=wasm_ref.id)
        assert wasm_ref == wasm_ref_2

        circuit = Circuit(1)
        a = circuit.add_c_register("a", 8)
        circuit.add_wasm_to_reg("add_one", qa_wasm_module, [a], [a])
        circuit.measure_all()
        qa_wasm_circuit_name_fixture = f"circuit with wasm for {test_case_name}"

        wasm_circuit_ref = qnx.circuits.upload(
            name=qa_wasm_circuit_name_fixture,
            circuit=circuit,
            project=proj_ref,
        )

        execute_job_ref = qnx.start_execute_job(
            programs=[wasm_circuit_ref],
            name=f"wasm execute job for {test_case_name}",
            n_shots=[100],
            backend_config=qnx.QuantinuumConfig(
                device_name="H1-Emulator",
            ),
            wasm_module=wasm_ref,
            project=proj_ref,
        )

        qnx.jobs.wait_for(execute_job_ref)
        assert qnx.jobs.status(execute_job_ref).status == JobStatusEnum.COMPLETED



"""Definitions of reference proxy objects to data in Nexus."""

from __future__ import annotations

from abc import abstractmethod
from copy import copy
from datetime import datetime
from enum import Enum
from typing import (
    Annotated,
    Any,
    Iterable,
    Literal,
    Optional,
    Protocol,
    TypeAlias,
    TypeVar,
    Union,
    cast,
)
from uuid import UUID

import pandas as pd
from hugr.package import Package
from hugr.qsystem.result import QsysResult
from pydantic import BaseModel, ConfigDict, Field, field_serializer
from pytket.backends.backendinfo import BackendInfo
from pytket.backends.backendresult import BackendResult
from pytket.circuit import Circuit
from pytket.wasm.wasm import WasmModuleHandler
from quantinuum_schemas.models.backend_config import BackendConfig

from qnexus.exceptions import IncompatibleResultVersion
from qnexus.models.annotations import Annotations
from qnexus.models.job_status import JobStatusEnum
from qnexus.models.utils import assert_never


class Dataframable(Protocol):
    """Protocol for structural subtyping of classes that
    have a default representation as a pandas.DataFrame."""

    @abstractmethod
    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        raise NotImplementedError


T = TypeVar("T", bound=Dataframable)


class DataframableList(list[T]):
    """A Python list that implements the Dataframable protocol."""

    def __init__(self, iterable: Iterable[T]) -> None:
        super().__init__(item for item in iterable)

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        if len(self) == 0:
            return pd.DataFrame()
        return pd.concat([item.df() for item in self], ignore_index=True)


class BaseRef(BaseModel):
    """Base pydantic model."""

    model_config = ConfigDict(frozen=True)

    id: UUID


class TeamRef(BaseRef):
    """Proxy object to a Team in Nexus."""

    name: str
    description: Optional[str]
    id: UUID
    type: Literal["TeamRef"] = "TeamRef"

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return pd.DataFrame(
            {
                "name": self.name,
                "description": self.description,
                "id": self.id,
            },
            index=[0],
        )


class UserRef(BaseRef):
    """Proxy object to a User in Nexus."""

    display_name: Optional[str]
    id: UUID
    type: Literal["UserRef"] = "UserRef"

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return pd.DataFrame(
            {
                "name": self.display_name,
                "id": self.id,
            },
            index=[0],
        )


class ProjectRef(BaseRef):
    """Proxy object to a Project in Nexus."""

    annotations: Annotations
    contents_modified: datetime
    archived: bool = Field(default=False)
    id: UUID
    type: Literal["ProjectRef"] = "ProjectRef"

    @field_serializer("contents_modified")
    def serialize_modified(self, contents_modified: datetime | None) -> str | None:
        """Custom serializer for datetimes."""
        if contents_modified:
            return str(contents_modified)
        return None

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return pd.DataFrame(
            {
                "name": self.annotations.name,
                "description": self.annotations.description,
                "created": self.annotations.created,
                "modified": self.annotations.modified,
                "contents_modified": self.contents_modified,
                "archived": self.archived,
                "id": self.id,
            },
            index=[0],
        )


class SystemRef(BaseRef):
    """Proxy object to a System in Nexus."""

    id: UUID
    name: str
    provider_name: str
    type: Literal["SystemRef"] = "SystemRef"

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return pd.DataFrame(
            {
                "id": self.id,
                "name": self.name,
                "provider_name": self.provider_name,
            },
            index=[0],
        )


class CircuitRef(BaseRef):
    """Proxy object to a Circuit in Nexus."""

    annotations: Annotations
    project: ProjectRef
    id: UUID
    _circuit: Circuit | None = None
    type: Literal["CircuitRef"] = "CircuitRef"

    def download_circuit(self) -> Circuit:
        """Get a copy of the circuit as a pytket ``Circuit`` object."""
        if self._circuit:
            return self._circuit.copy()
        from qnexus.client.circuits import _fetch_circuit

        self._circuit = _fetch_circuit(self)
        return self._circuit.copy()

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return self.annotations.df().join(
            pd.DataFrame(
                {
                    "project": self.project.annotations.name,
                    "id": self.id,
                },
                index=[0],
            )
        )


class WasmModuleRef(BaseRef):
    """Proxy object to a WasmModule in Nexus."""

    annotations: Annotations
    project: ProjectRef
    id: UUID
    _contents: WasmModuleHandler | None = None
    type: Literal["WasmModuleRef"] = "WasmModuleRef"

    def download_wasm_contents(self) -> WasmModuleHandler:
        """Get the contents of the original uploaded WASM."""
        if self._contents:
            return self._contents

        from qnexus.client.wasm_modules import _fetch_wasm_module

        self._contents = _fetch_wasm_module(self)
        return self._contents

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return self.annotations.df().join(
            pd.DataFrame(
                {
                    "project": self.project.annotations.name,
                    "id": self.id,
                },
                index=[0],
            )
        )


class GpuDecoderConfigRef(BaseRef):
    """Proxy object to a GpuDecoderConfig in Nexus."""

    annotations: Annotations
    project: ProjectRef
    id: UUID
    _contents: str | None = None
    type: Literal["GpuDecoderConfigRef"] = "GpuDecoderConfigRef"

    def download_gpu_decoder_config_contents(self) -> str:
        """Get the contents of the original uploaded gpu decoder config."""
        if self._contents:
            return self._contents

        from qnexus.client.gpu_decoder_configs import (
            _fetch_gpu_decoder_config,
        )

        self._contents = _fetch_gpu_decoder_config(self)
        return self._contents

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return self.annotations.df().join(
            pd.DataFrame(
                {
                    "project": self.project.annotations.name,
                    "id": self.id,
                },
                index=[0],
            )
        )


class HUGRRef(BaseRef):
    """Proxy object to a HUGR in Nexus."""

    annotations: Annotations
    project: ProjectRef
    id: UUID
    _contents: Package | None = None
    _bytes: bytes | None = None
    type: Literal["HUGRRef"] = "HUGRRef"

    def download_hugr(self) -> Package:
        """Get the HUGR Package of the original uploaded HUGR."""

        if self._contents:
            return self._contents

        from qnexus.client.hugr import _fetch_hugr_package

        self._contents = _fetch_hugr_package(self)
        return self._contents

    def download_hugr_bytes(self) -> bytes:
        """Get the HUGR bytes of the original uploaded HUGR."""

        if self._bytes:
            return self._bytes

        from qnexus.client.hugr import _fetch_hugr_bytes

        self._bytes = _fetch_hugr_bytes(self)
        return self._bytes

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return self.annotations.df().join(
            pd.DataFrame(
                {
                    "project": self.project.annotations.name,
                    "id": self.id,
                },
                index=[0],
            )
        )


class QIRRef(BaseRef):
    """Proxy object to a QIR program in Nexus."""

    annotations: Annotations
    project: ProjectRef
    id: UUID
    _contents: bytes | None = None
    type: Literal["QIRRef"] = "QIRRef"

    def download_qir(self) -> bytes:
        """Get the QIR program."""

        if self._contents:
            return self._contents

        from qnexus.client.qir import _fetch_qir

        self._contents = _fetch_qir(self)
        return self._contents

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return self.annotations.df().join(
            pd.DataFrame(
                {
                    "project": self.project.annotations.name,
                    "id": self.id,
                },
                index=[0],
            )
        )


class JobType(str, Enum):
    """Enum for a job's type."""

    EXECUTE = "execute"
    COMPILE = "compile"


class JobRef(BaseRef):
    """Proxy object to a Job in Nexus."""

    model_config = ConfigDict(frozen=False)

    annotations: Annotations
    job_type: JobType
    last_status: JobStatusEnum
    last_message: str
    project: ProjectRef
    system: SystemRef | None = None
    id: UUID
    backend_config_store: BackendConfig | None = None
    type: Literal["JobRef", "CompileJobRef", "ExecuteJobRef"] = "JobRef"

    @property
    def backend_config(self) -> BackendConfig:
        """Fetch the backend_config for a job."""
        from qnexus.client.jobs import _fetch_by_id

        if self.backend_config_store:
            return self.backend_config_store
        self.backend_config_store = cast(
            BackendConfig, _fetch_by_id(self.id, None).backend_config_store
        )
        return self.backend_config_store

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return self.annotations.df().join(
            pd.DataFrame(
                {
                    "job_type": self.job_type,
                    "last_status": self.last_status,
                    "project": self.project.annotations.name,
                    "backend_config": self.backend_config.__class__.__name__,
                    "system": self.system.name if self.system else "Unknown",
                    "id": self.id,
                },
                index=[0],
            )
        )


class CompileJobRef(JobRef, BaseRef):
    """Proxy object to a CompileJob in Nexus."""

    model_config = ConfigDict(frozen=False)

    job_type: JobType = JobType.COMPILE
    type: Literal["CompileJobRef"] = "CompileJobRef"


class ExecuteJobRef(JobRef, BaseRef):
    """Proxy object to an ExecuteJob in Nexus."""

    model_config = ConfigDict(frozen=False)

    job_type: JobType = JobType.EXECUTE
    type: Literal["ExecuteJobRef"] = "ExecuteJobRef"


class CompilationResultRef(BaseRef):
    """Proxy object to the results of a circuit compilation in Nexus."""

    annotations: Annotations
    project: ProjectRef
    _input_circuit: CircuitRef | None = None
    _output_circuit: CircuitRef | None = None
    _compilation_passes: DataframableList[CompilationPassRef] | None = None
    id: UUID  # compilation id
    type: Literal["CompilationResultRef"] = "CompilationResultRef"

    def get_input(self) -> CircuitRef:
        """Get the CircuitRef of the original circuit."""
        if self._input_circuit:
            return self._input_circuit

        from qnexus.client.jobs._compile import _fetch_compilation_output

        (self._input_circuit, self._output_circuit) = _fetch_compilation_output(self)
        return self._input_circuit

    def get_output(self) -> CircuitRef:
        """Get the CircuitRef of the compiled circuit."""
        if self._output_circuit:
            return self._output_circuit

        from qnexus.client.jobs._compile import _fetch_compilation_output

        (self._input_circuit, self._output_circuit) = _fetch_compilation_output(self)
        return self._output_circuit

    def get_passes(self) -> DataframableList[CompilationPassRef]:
        """Get information on the compilation passes and the output circuits (if available)."""
        if self._compilation_passes:
            return copy(self._compilation_passes)

        self._compilation_passes = self._get_compile_results()
        return copy(self._compilation_passes)

    def _get_compile_results(
        self,
    ) -> DataframableList[CompilationPassRef]:
        """Utility method to retrieve the passes and output circuit."""
        from qnexus.client.jobs._compile import _fetch_compilation_passes

        passes = _fetch_compilation_passes(self)
        return passes

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return self.annotations.df().join(
            pd.DataFrame(
                {
                    "project": self.project.annotations.name,
                    "id": self.id,
                },
                index=[0],
            )
        )


class ResultType(str, Enum):
    """Enum for a results's type."""

    PYTKET = "PYTKET"
    QSYS = "QSYS"


class ResultVersions(int, Enum):
    """Enumerate the valid values for requesting results in a specific format"""

    DEFAULT = 3
    RAW = 4


class QIRResult:
    results: str

    def __init__(self, results: str):
        self.results = results


ExecutionProgram: TypeAlias = CircuitRef | HUGRRef | QIRRef
ExecutionResult: TypeAlias = QsysResult | BackendResult | QIRResult


class ExecutionResultRef(BaseRef):
    """Proxy object to the results of a circuit execution through Nexus."""

    annotations: Annotations
    project: ProjectRef
    result_type: ResultType = ResultType.PYTKET
    _input_program: ExecutionProgram | None = None
    _result: ExecutionResult | None = None
    _result_version: ResultVersions | None = None
    _backend_info: BackendInfo | None = None
    id: UUID
    type: Literal["ExecutionResultRef"] = "ExecutionResultRef"

    def get_input(self) -> ExecutionProgram:
        """Get the Program Ref of the input program."""
        if self._input_program:
            return self._input_program

        (
            self._result,
            self._backend_info,
            self._input_program,
        ) = self._get_execute_results(ResultVersions.DEFAULT)
        self._result_version = ResultVersions.DEFAULT
        return copy(self._input_program)

    def download_result(
        self, version: ResultVersions = ResultVersions.DEFAULT
    ) -> ExecutionResult:
        """Get a copy of the result of the program execution."""
        if self._result and self._result_version == version:
            return copy(self._result)

        (
            self._result,
            self._backend_info,
            self._input_program,
        ) = self._get_execute_results(version=version)
        self._result_version = version
        return copy(self._result)

    def download_backend_info(self) -> BackendInfo:
        """Get a copy of the pytket BackendInfo."""
        if self._backend_info:
            return copy(self._backend_info)

        (
            self._result,
            self._backend_info,
            self._input_program,
        ) = self._get_execute_results(ResultVersions.DEFAULT)
        self._result_version = ResultVersions.DEFAULT
        return copy(self._backend_info)

    def _get_execute_results(
        self, version: ResultVersions
    ) -> tuple[ExecutionResult, BackendInfo, ExecutionProgram]:
        """Utility method to retrieve the passes and output circuit.
        result_version can be passed to request v4 results for qsys results only.

        Default results for any program type on H series devices is pytket style results
        Default result for QIR programs on NG devices is QIR standard compliant results.
        Default result for HUGR programs is NG results.
        """
        from qnexus.client.jobs._execute import (
            _fetch_pytket_execution_result,
            _fetch_qsys_execution_result,
        )

        match self.result_type:
            case ResultType.PYTKET:
                if version != ResultVersions.DEFAULT:
                    raise IncompatibleResultVersion(
                        "pytket results can only be fetched in the default version"
                    )
                return _fetch_pytket_execution_result(self)
            case ResultType.QSYS:
                return _fetch_qsys_execution_result(self, version)
            case _:
                assert_never(self.result_type)

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return self.annotations.df().join(
            pd.DataFrame(
                {
                    "project": self.project.annotations.name,
                    "id": self.id,
                    "result_type": self.result_type,
                },
                index=[0],
            )
        )


class CompilationPassRef(BaseRef):
    """Proxy object to a compilation pass that was applied on a circuit in Nexus."""

    pass_name: str
    input_circuit: CircuitRef
    output_circuit: CircuitRef
    id: UUID
    type: Literal["CompilationPassRef"] = "CompilationPassRef"

    def get_input(self) -> CircuitRef:
        """Get the CircuitRef of the original circuit."""
        return self.input_circuit

    def get_output(self) -> CircuitRef:
        """Get the CircuitRef of the compiled circuit."""
        return self.output_circuit

    def df(self) -> pd.DataFrame:
        """Present in a pandas DataFrame."""
        return pd.DataFrame(
            {
                "pass name": self.pass_name,
                "input": self.input_circuit.annotations.name,
                "output": self.output_circuit.annotations.name,
                "id": self.id,
            },
            index=[0],
        )


Ref = Annotated[
    Union[
        TeamRef,
        UserRef,
        ProjectRef,
        CircuitRef,
        WasmModuleRef,
        GpuDecoderConfigRef,
        HUGRRef,
        QIRRef,
        JobRef,
        CompileJobRef,
        ExecuteJobRef,
        CompilationResultRef,
        ExecutionResultRef,
        CompilationPassRef,
        SystemRef,
    ],
    Field(discriminator="type"),
]


ref_name_to_class: dict[str, Ref] = {
    config_type.__name__: config_type  # type: ignore
    for config_type in BaseRef.__subclasses__()
}


def deserialize_nexus_ref(jsonable: dict[str, Any]) -> Ref:
    """Deserialize something that should be a subclass of BaseRef based on
    the value of its 'type' field."""
    ref_type = jsonable["type"]
    if ref_type in ref_name_to_class.keys():
        ref_class = ref_name_to_class[ref_type]
        return ref_class(**jsonable)  # type: ignore
    raise ValueError(
        f"Cannot deserialize as {ref_type}, no known class matches that value."
    )



"""Language options for remote job submission."""

from enum import Enum


class Language(str, Enum):
    """Enumeration for the possible submission languages for remote submissions."""

    AUTO = "AUTO"
    QASM = "OPENQASM 2.0"
    QIR = "QIR 1.0"



"""Test basic functionality relating to the quota module."""

import pandas as pd

import qnexus as qnx
from qnexus.models import Quota


def test_quota_get(
    authenticated_nexus: None,
) -> None:
    """Test that we can get a specific Quota."""

    my_quota = qnx.quotas.get(name="compilation")
    assert isinstance(my_quota, Quota)


def test_quota_get_all(
    authenticated_nexus: None,
) -> None:
    """Test that we can get all quotas (currently not a NexusIterator)."""

    my_quotas = qnx.quotas.get_all()

    assert len(my_quotas) > 1
    assert isinstance(my_quotas.df(), pd.DataFrame)
    assert isinstance(my_quotas[0], Quota)


def test_check_quota(
    authenticated_nexus: None,
) -> None:
    """Test that we can check a specific quota guard."""

    # Assumes the user isn't over their Jupyterhub quota
    assert qnx.quotas.check_quota(name="jupyterhub") is True



import typing
import warnings
from importlib.metadata import version
from unittest import mock

import httpx
import respx

import qnexus as qnx
from qnexus.client import (
    LATEST_VERSION_HEADER,
    VERSION_HEADER,
    VERSION_STATUS_HEADER,
    get_nexus_client,
)
from qnexus.client.utils import write_token

FAKE_LATEST_VERSION = "999.99.999-never-gonna-happen"
FAKE_VERSION_STATUS = "really bad"


@respx.mock
def test_sunset_header_emits_warning() -> None:
    fake_date = "foo"
    path = "/api/projects/v1beta2"
    list_project_route = respx.get(f"{get_nexus_client().base_url}{path}").mock(
        return_value=httpx.Response(
            status_code=200,
            json={"data": {}},
            headers={"Sunset": fake_date},
        )
    )

    with warnings.catch_warnings(record=True) as captured:
        qnx.projects.get_all().list()

    assert list_project_route.called
    assert len(captured) == 1
    assert captured[0].category is DeprecationWarning
    message = str(captured[0].message)
    assert fake_date in message
    assert "(" + path + ")" in message


def _check_version_warning_emitted(warning_msgs: list[warnings.WarningMessage]) -> None:
    """
    Assert that a warning has been emitted with all the properties we'd expect
    when we notify a user that they need to upgrade their client.
    """
    for warning_msg in warning_msgs:
        try:
            assert warning_msg.category is DeprecationWarning
            message = str(warning_msg.message)
            assert version("qnexus") in message
            assert FAKE_LATEST_VERSION in message
            assert FAKE_VERSION_STATUS in message
            assert "Please consider upgrading" in message
            return
        except AssertionError:
            pass  # on to the next warning, if any
    assert False, (
        f"The expected warning was not found (checked {len(warning_msgs)} warning messages)."
    )


def _check_request_includes_version_data(r: respx.Route) -> None:
    """
    Verify that a single call to a server URL has been made, and that it
    includes the expected version identifier header.
    """
    assert r.call_count == 1
    call: respx.models.Call = r.calls[0]
    headers: typing.MutableMapping[str, str] = call.request.headers
    assert headers[VERSION_HEADER] == version("qnexus")


@respx.mock
def test_version_check_emits_warning_refresh_token() -> None:
    """
    GIVEN: a server that includes a http header with a new version of qnexus
           and an indication that upgrade is advised
    WHEN:  the auth token is refreshed from an old version of qnexus
    THEN:  a warning is emitted, prompting the user to upgrade
    """
    write_token("refresh_token", "dummy_oat")

    # Mock the list projects endpoint to force a token refresh
    respx.get(f"{get_nexus_client().base_url}/api/projects/v1beta2").mock(
        side_effect=[
            httpx.Response(401),
            httpx.Response(200, json={"included": {}, "data": []}),
        ]
    )

    # Mock the token refresh endpoint
    refresh_token_route = respx.post(
        f"{get_nexus_client().base_url}/auth/tokens/refresh"
    ).mock(
        return_value=httpx.Response(
            status_code=200,
            headers={
                LATEST_VERSION_HEADER: FAKE_LATEST_VERSION,
                VERSION_STATUS_HEADER: "x;" + FAKE_VERSION_STATUS,
            },
        )
    )

    with warnings.catch_warnings(record=True) as captured:
        qnx.projects.get_all().list()

    _check_request_includes_version_data(refresh_token_route)
    _check_version_warning_emitted(captured)


@mock.patch("qnexus.client.auth.webbrowser")
@respx.mock
def test_version_check_emits_warning_login(m: mock.Mock) -> None:
    """
    GIVEN: a server that includes a http header with a new version of qnexus
           and an indication that upgrade is advised
    WHEN:  the login method is called from an old version of qnexus
    THEN:  a warning is emitted, prompting the user to upgrade
    """
    base_url = get_nexus_client().base_url

    # Mock the login endpoints
    respx.post(f"{base_url}/auth/device/device_authorization").mock(
        return_value=httpx.Response(
            status_code=200,
            json={
                "user_code": "",
                "device_code": "",
                "verification_uri_complete": "https://example.com",
                "expires_in": 5,
                "interval": 1,
            },
        )
    )
    token_request_route = respx.post(f"{base_url}/auth/device/token").mock(
        return_value=httpx.Response(
            status_code=200,
            json={
                "refresh_token": "foo",
                "access_token": "bar",
                "email": "foo@example.com",
            },
            headers={
                LATEST_VERSION_HEADER: FAKE_LATEST_VERSION,
                VERSION_STATUS_HEADER: "x;" + FAKE_VERSION_STATUS,
            },
        )
    )

    with warnings.catch_warnings(record=True) as captured:
        qnx.login()

    _check_request_includes_version_data(token_request_route)
    _check_version_warning_emitted(captured)



"""Tests that we can use all available backend configs."""

from typing import Callable, ContextManager

from pytket.backends.backendresult import BackendResult
from pytket.circuit import Circuit

import qnexus as qnx
from qnexus.models.job_status import JobStatusEnum
from qnexus.models.references import ProjectRef

CONFIGS_REQUIRE_NO_MEASURE = [qnx.AerUnitaryConfig]
CONFIGS_NOT_TO_EXECUTE = [
    # We won't execute for these configs as they run on quantum hardware
    qnx.IBMQConfig
]


def test_basic_backend_config_usage(
    backend_config: qnx.BackendConfig,
    create_project: Callable[[str], ContextManager[ProjectRef]],
    test_case_name: str,
) -> None:
    """Test basic functionality of supported BackendConfigs."""

    with create_project(f"project for {test_case_name}") as project_ref:
        my_circ = Circuit(2, 2).H(0).CX(0, 1)

        if backend_config.__class__ not in CONFIGS_REQUIRE_NO_MEASURE:
            my_circ.measure_all()

        my_circ = qnx.circuits.upload(
            circuit=my_circ,
            name=f"circuit for {test_case_name}",
            description="This can be safely deleted.",
            project=project_ref,
        )

        compile_job_ref = qnx.start_compile_job(
            programs=[my_circ],
            name=f"compile job for {test_case_name}",
            optimisation_level=2,
            backend_config=backend_config,
            project=project_ref,
        )

        qnx.jobs.wait_for(compile_job_ref)

        # Check the backend config as stored in Nexus matches the one specified
        # QuantinuumConfig noisy_simulation flag getting reset in Nexus
        assert backend_config.model_dump(exclude={"noisy_simulation"}) == qnx.jobs.get(
            id=compile_job_ref.id
        ).backend_config.model_dump(exclude={"noisy_simulation"})

        execute_job_ref = qnx.start_execute_job(
            programs=[item.get_output() for item in qnx.jobs.results(compile_job_ref)],
            name=f"execute job for {test_case_name}",
            n_shots=[100],
            backend_config=backend_config,
            project=project_ref,
        )

        if backend_config.__class__ in CONFIGS_NOT_TO_EXECUTE:
            qnx.jobs.wait_for(execute_job_ref, wait_for_status=JobStatusEnum.SUBMITTED)
            qnx.jobs.cancel(execute_job_ref)
            qnx.jobs.wait_for(execute_job_ref, wait_for_status=JobStatusEnum.CANCELLED)
            return

        qnx.jobs.wait_for(execute_job_ref)

        execute_job_result_refs = qnx.jobs.results(execute_job_ref)

        for result_ref in execute_job_result_refs:
            assert isinstance(result_ref.download_result(), BackendResult)



"""Test basic functionality relating to the project module."""

from datetime import datetime
from typing import Callable, ContextManager

import pandas as pd
import pytest
from pytket.circuit import Circuit

import qnexus as qnx
import qnexus.exceptions as qnx_exc
from qnexus.models.references import CompileJobRef, ProjectRef, Ref


def test_project_get(
    test_case_name: str,
    create_project: Callable[[str], ContextManager[ProjectRef]],
    test_ref_serialisation: Callable[[str, Ref], None],
) -> None:
    """Test that we can get a specific unique project
    by name or id, or get an appropriate exception."""

    with create_project(f"project for {test_case_name}") as proj_ref:
        assert isinstance(proj_ref, ProjectRef)

        my_proj_2 = qnx.projects.get(id=proj_ref.id)
        assert proj_ref == my_proj_2

        with pytest.raises(qnx_exc.NoUniqueMatch):
            qnx.projects.get()

        with pytest.raises(qnx_exc.ZeroMatches):
            qnx.projects.get(name_like=f"{datetime.now()}_{datetime.now()}")

        test_ref_serialisation("project", my_proj_2)


def test_project_get_all(authenticated_nexus: None) -> None:
    """Test that we can get an iterator over all projects."""

    my_proj_db_matches = qnx.projects.get_all()

    assert isinstance(my_proj_db_matches.count(), int)
    assert isinstance(my_proj_db_matches.summarize(), pd.DataFrame)

    assert isinstance(next(my_proj_db_matches), ProjectRef)


def test_project_create(
    test_case_name: str,
    authenticated_nexus: None,
) -> None:
    """Test that we can create a project and add a property definition."""

    project_name = f"project for {test_case_name}"

    my_new_project = qnx.projects.create(name=project_name)

    assert isinstance(my_new_project, ProjectRef)

    test_property_name = f"property for {test_case_name}"

    qnx.projects.add_property(
        name=test_property_name,
        property_type="bool",
        project=my_new_project,
        description="A test property for my QA test project",
    )

    test_props = qnx.projects.get_properties(my_new_project)

    assert len(test_props) == 1
    assert test_props[0].annotations.name == test_property_name

    with pytest.raises(qnx_exc.ResourceDeleteFailed):
        qnx.projects.delete(my_new_project)

    qnx.projects.update(my_new_project, archive=True)

    qnx.projects.delete(my_new_project)

    with pytest.raises(qnx_exc.ZeroMatches):
        qnx.projects.get(name_like=project_name)


def test_project_get_or_create(
    test_case_name: str,
    authenticated_nexus: None,
) -> None:
    """Test that we can get or create a project."""

    project_name = f"project for {test_case_name}"

    with pytest.raises(qnx_exc.ZeroMatches):
        qnx.projects.get(name_like=project_name)

    my_new_project = qnx.projects.get_or_create(name=project_name)

    assert isinstance(my_new_project, ProjectRef)

    my_new_project_again = qnx.projects.get_or_create(name=project_name)

    assert my_new_project == my_new_project_again

    qnx.projects.update(my_new_project, archive=True)
    qnx.projects.delete(my_new_project)


def test_project_summarize(
    test_case_name: str,
    create_compile_job_in_project: Callable[..., ContextManager[CompileJobRef]],
) -> None:
    """Test that we can get summary information on the state of a project."""
    project_name = f"project for {test_case_name}"
    circuit_name = f"circuit for {test_case_name}"
    compile_job_name = f"compile job for {test_case_name}"

    with create_compile_job_in_project(
        project_name=project_name,
        job_name=compile_job_name,
        circuit=Circuit(2, 2).H(0).CX(0, 1).measure_all(),
        circuit_name=circuit_name,
    ):
        my_proj = qnx.projects.get(name_like=project_name)
        assert isinstance(my_proj, ProjectRef)

        project_summary = qnx.projects.summarize(my_proj)

        assert isinstance(project_summary, pd.DataFrame)

        assert (project_summary["total_jobs"] > 0).all()



"""Test basic functionality relating to the role module."""

from typing import Callable, ContextManager

import pandas as pd

import qnexus as qnx
from qnexus.config import CONFIG
from qnexus.models import Role
from qnexus.models.references import ProjectRef, TeamRef, UserRef


def test_role_get(authenticated_nexus: None) -> None:
    """Test that we can get a specific Role."""

    role = qnx.roles.get(name="Administrator")
    assert isinstance(role, Role)


def test_role_get_all(authenticated_nexus: None) -> None:
    """Test that we can get all assignment role definitions."""

    all_roles = qnx.roles.get_all()

    assert len(all_roles) == 4
    assert isinstance(all_roles.df(), pd.DataFrame)
    assert all([isinstance(role, Role) for role in list(all_roles)])


def test_team_assignment(
    test_case_name: str,
    create_project: Callable[[str], ContextManager[ProjectRef]],
    create_team: Callable[[str], ContextManager[TeamRef]],
) -> None:
    """Test that we can assign a team to a project."""

    # Set up
    team_name = f"{test_case_name[-86:]}"  # TODO: use full name once bug is fixed
    project_name = f"project for {test_case_name}"

    with create_team(team_name) as team_ref:
        with create_project(project_name) as proj_ref:
            qnx.roles.assign_team(
                resource_ref=proj_ref, team=team_ref, role="Administrator"
            )
            assignments = qnx.roles.assignments(resource_ref=proj_ref)
            assert len(assignments) == 2

            team_assignment = next(
                assign for assign in assignments if assign.assignment_type == "team"
            )
            assert team_assignment.assignment_type == "team"
            assert isinstance(team_assignment.assignee, TeamRef)
            assert team_assignment.assignee.id == team_ref.id


def test_user_assignment(
    test_case_name: str, create_project: Callable[[str], ContextManager[ProjectRef]]
) -> None:
    """Test that we can assign a role to a user."""
    with create_project(f"project for {test_case_name}") as new_project_ref:
        qnx.roles.assign_user(
            resource_ref=new_project_ref,
            user_email=CONFIG.qa_user_email,
            role="Contributor",
        )

        assignments = qnx.roles.assignments(resource_ref=new_project_ref)

        assert len(assignments) == 2

        contrib_assignment = next(
            assign for assign in assignments if assign.role.name == "Contributor"
        )
        assert contrib_assignment.assignment_type == "user"
        assert isinstance(contrib_assignment.assignee, UserRef)

        admin_assignment = next(
            assign for assign in assignments if assign.role.name == "Administrator"
        )
        assert admin_assignment.assignment_type == "user"
        assert isinstance(admin_assignment.assignee, UserRef)



"""CLI for qnexus."""

from typing import Any

import click

from qnexus.client.jobs import Params, get_all

from .utils import add_options_to_command, is_documented_by


@click.command()
@is_documented_by(get_all)
def jobs(**kwargs: Any) -> None:
    """List jobs."""
    click.echo(get_all(**kwargs))


add_options_to_command(jobs, Params)



"""Basic checks for HUGR functionality."""

import uuid
from datetime import datetime

import pytest

from qnexus.exceptions import IncompatibleResultVersion
from qnexus.models.annotations import Annotations
from qnexus.models.references import ExecutionResultRef, ProjectRef, ResultVersions


def test_raises_when_trying_to_get_raw_results_from_pytket_result() -> None:
    """Given an ExecutionResultRef to a pytket result, it will raise an error
    if the user tries to get anything other than the default results."""

    ref = ExecutionResultRef(
        id=uuid.uuid4(),
        annotations=Annotations(),
        project=ProjectRef(
            id=uuid.uuid4(), annotations=Annotations(), contents_modified=datetime.now()
        ),
    )

    with pytest.raises(IncompatibleResultVersion):
        ref.download_result(version=ResultVersions.RAW)



"""Client API for execution in Nexus."""

from typing import Union, cast

from hugr.qsystem.result import QsysResult
from pytket.backends.backendinfo import BackendInfo
from pytket.backends.backendresult import BackendResult

import qnexus.exceptions as qnx_exc
from qnexus.client import circuits as circuit_api
from qnexus.client import get_nexus_client
from qnexus.client import hugr as hugr_api
from qnexus.client import qir as qir_api
from qnexus.client.utils import accept_circuits_for_programs
from qnexus.context import get_active_project, merge_properties_from_context
from qnexus.models import BackendConfig, StoredBackendInfo, to_pytket_backend_info
from qnexus.models.annotations import Annotations, CreateAnnotations, PropertiesDict
from qnexus.models.job_status import JobStatusEnum
from qnexus.models.language import Language
from qnexus.models.references import (
    CircuitRef,
    DataframableList,
    ExecuteJobRef,
    ExecutionProgram,
    ExecutionResultRef,
    GpuDecoderConfigRef,
    HUGRRef,
    JobType,
    ProjectRef,
    QIRRef,
    QIRResult,
    ResultType,
    ResultVersions,
    WasmModuleRef,
)
from qnexus.models.utils import assert_never


@accept_circuits_for_programs
@merge_properties_from_context
def start_execute_job(
    programs: Union[ExecutionProgram, list[ExecutionProgram]],
    n_shots: list[int] | list[None],
    backend_config: BackendConfig,
    name: str,
    description: str = "",
    properties: PropertiesDict | None = None,
    project: ProjectRef | None = None,
    valid_check: bool = True,
    language: Language = Language.AUTO,
    credential_name: str | None = None,
    wasm_module: WasmModuleRef | None = None,
    gpu_decoder_config: GpuDecoderConfigRef | None = None,
    user_group: str | None = None,
) -> ExecuteJobRef:
    """
    Submit an execute job to be run in Nexus. Returns an ``ExecuteJobRef``
    object which can be used to check the job's status.  See ``qnexus.execute``
    for a utility method that waits for the results and returns them.
    """
    project = project or get_active_project(project_required=True)
    project = cast(ProjectRef, project)

    program_ids = (
        [str(p.id) for p in programs]
        if isinstance(programs, list)
        else [str(programs.id)]
    )

    if len(n_shots) != len(program_ids):
        raise ValueError("Number of programs must equal number of n_shots.")

    attributes_dict = CreateAnnotations(
        name=name,
        description=description,
        properties=properties,
    ).model_dump(exclude_none=True)
    attributes_dict.update(
        {
            "job_type": "execute",
            "definition": {
                "job_definition_type": "execute_job_definition",
                "backend_config": backend_config.model_dump(exclude_none=True),
                "user_group": user_group,
                "valid_check": valid_check,
                "language": (
                    language.value if isinstance(language, Language) else language
                ),
                "wasm_module_id": str(wasm_module.id) if wasm_module else None,
                "gpu_decoder_config_id": (
                    str(gpu_decoder_config.id) if gpu_decoder_config else None
                ),
                "credential_name": credential_name,
                "items": [
                    {"program_id": program_id, "n_shots": n_shot}
                    for program_id, n_shot in zip(program_ids, n_shots)
                ],
            },
        }
    )
    relationships = {
        "project": {"data": {"id": str(project.id), "type": "project"}},
    }
    req_dict = {
        "data": {
            "attributes": attributes_dict,
            "relationships": relationships,
            "type": "job",
        }
    }

    resp = get_nexus_client().post(
        "/api/jobs/v1beta3",
        json=req_dict,
    )
    if resp.status_code != 202:
        raise qnx_exc.ResourceCreateFailed(
            message=resp.text, status_code=resp.status_code
        )

    return ExecuteJobRef(
        id=resp.json()["data"]["id"],
        annotations=Annotations.from_dict(resp.json()["data"]["attributes"]),
        job_type=JobType.EXECUTE,
        last_status=JobStatusEnum.SUBMITTED,
        last_message="",
        project=project,
        backend_config_store=backend_config,
    )


def _results(
    execute_job: ExecuteJobRef,
    allow_incomplete: bool = False,
) -> DataframableList[ExecutionResultRef]:
    """Get the results from an execute job."""

    resp = get_nexus_client().get(f"/api/jobs/v1beta3/{execute_job.id}")
    if resp.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(
            message=resp.text, status_code=resp.status_code
        )
    resp_data = resp.json()["data"]
    job_status = resp_data["attributes"]["status"]["status"]

    if job_status != "COMPLETED" and not allow_incomplete:
        raise qnx_exc.ResourceFetchFailed(message=f"Job status: {job_status}")

    execute_results: DataframableList[ExecutionResultRef] = DataframableList([])

    for item in resp_data["attributes"]["definition"]["items"]:
        # Check if item is in a state that returns results
        if item["status"]["status"] in (
            "CANCELLED",
            "ERROR",
            "DEPLETED",
            "TERMINATED",
            "COMPLETED",
            "RUNNING",
        ):
            result_type: ResultType

            match item.get("result_type", None):
                case ResultType.QSYS:
                    result_type = ResultType.QSYS
                case ResultType.PYTKET:
                    result_type = ResultType.PYTKET
                case None:
                    continue
                case _:
                    assert_never(item["result_type"])

            result_ref = ExecutionResultRef(
                id=item["result_id"],
                annotations=execute_job.annotations,
                project=execute_job.project,
                result_type=result_type,
            )

            execute_results.append(result_ref)

    return execute_results


def _fetch_pytket_execution_result(
    result_ref: ExecutionResultRef,
) -> tuple[BackendResult, BackendInfo, Union[CircuitRef, QIRRef]]:
    """Get the results for an execute job item."""
    assert result_ref.result_type == ResultType.PYTKET, "Incorrect result type"

    res = get_nexus_client().get(f"/api/results/v1beta3/{result_ref.id}")
    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(message=res.text, status_code=res.status_code)

    res_dict = res.json()
    program_data = res_dict["data"]["relationships"]["program"]["data"]
    program_id = program_data["id"]
    program_type = program_data["type"]

    input_program: Union[CircuitRef | QIRRef]
    match program_type:
        case "circuit":
            input_program = circuit_api._fetch_by_id(program_id, scope=None)
        case "qir":
            input_program = qir_api._fetch_by_id(program_id, scope=None)
        case _:
            raise ValueError(f"Unknown program type {type}")

    results_data = res_dict["data"]["attributes"]

    results_dict = {k: v for k, v in results_data.items() if v != [] and v is not None}

    backend_result = BackendResult.from_dict(results_dict)

    backend_info_data = next(
        data for data in res_dict["included"] if data["type"] == "backend_snapshot"
    )
    backend_info = to_pytket_backend_info(
        StoredBackendInfo(**backend_info_data["attributes"])
    )

    return (backend_result, backend_info, input_program)


def _fetch_qsys_execution_result(
    result_ref: ExecutionResultRef,
    version: ResultVersions,
) -> tuple[QsysResult | QIRResult, BackendInfo, HUGRRef | QIRRef]:
    """Get the results of a next-gen Qsys execute job."""
    assert result_ref.result_type == ResultType.QSYS, "Incorrect result type"

    chunk_number = 0
    params = {"version": version.value, "chunk": chunk_number}

    res = get_nexus_client().get(
        f"/api/qsys_results/v1beta2/partial/{result_ref.id}", params=params
    )

    if res.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(message=res.text, status_code=res.status_code)

    # This is only needed to be set once, as subsequent calls will
    # return the same information for the relationships.
    res_dict = res.json()
    input_program_id = res_dict["data"]["relationships"]["program"]["data"]["id"]

    input_program: HUGRRef | QIRRef
    result: QsysResult | QIRResult
    match res_dict["data"]["relationships"]["program"]["data"]["type"]:
        case "hugr":
            input_program = hugr_api._fetch_by_id(
                input_program_id,
                scope=None,
            )
            result = QsysResult(res_dict["data"]["attributes"].get("results"))
        case "qir":
            input_program = qir_api._fetch_by_id(
                input_program_id,
                scope=None,
            )
            if version == ResultVersions.DEFAULT:
                result = QIRResult(res_dict["data"]["attributes"].get("results"))
            else:
                result = QsysResult(res_dict["data"]["attributes"].get("results"))

    backend_info_data = next(
        data for data in res_dict["included"] if data["type"] == "backend_snapshot"
    )
    backend_info = to_pytket_backend_info(
        StoredBackendInfo(**backend_info_data["attributes"])
    )

    # We shouldn't be doing infinite loops, but the API currently doesn't
    # provide a way to know how many chunks there are, so we loop until we
    # get all of them.
    while True:
        chunk_number += 1
        params["chunk_number"] = chunk_number
        partial = get_nexus_client().get(
            f"/api/qsys_results/v1beta2/partial/{result_ref.id}", params=params
        )
        if partial.status_code == 404:
            # No more chunks. Stop here.
            break
        if partial.status_code != 200:
            raise qnx_exc.ResourceFetchFailed(
                message=res.text, status_code=partial.status_code
            )
        if isinstance(result.results, str):
            assert (
                version == ResultVersions.DEFAULT
            )  # Only QIR outputs are in this mode
            prev_str = result.results.split("END")[
                0
            ]  # remove the end tag from result.results
            next_str = "\n".join(
                [
                    line
                    for line in QIRResult(
                        partial.json()["data"]["attributes"]["results"]
                    ).results.splitlines()
                    if "OUTPUT" in line
                ]
            )  # just the output lines
            result.results += (
                prev_str + next_str + "END\t0\n"
            )  # join everything back up
        else:
            next_res = QsysResult(partial.json()["data"]["attributes"]["results"])
            result.results.extend(next_res.results)

    return (
        result,
        backend_info,
        input_program,
    )



"""Test basic functionality relating to the circuit module."""

from typing import Callable, ContextManager

import pandas as pd
import pytest
from pytket.circuit import Circuit

import qnexus as qnx
import qnexus.exceptions as qnx_exc
from qnexus.models.annotations import PropertiesDict
from qnexus.models.references import CircuitRef, ProjectRef, Ref

test_circuit = Circuit(2, 2).H(0).CX(0, 1).measure_all()


def test_circuit_get(
    test_case_name: str,
    create_circuit_in_project: Callable[
        [Circuit, str, str], ContextManager[CircuitRef]
    ],
    test_ref_serialisation: Callable[[str, Ref], None],
) -> None:
    """Test that we can create a circuit, add a property value,
    get a specific unique CircuitRef by name or id,  get an iterator over all circuits,
    and the serialisation round trip of a CircuitRef.
    """

    with create_circuit_in_project(
        test_circuit,
        f"project for {test_case_name}",
        f"circuit for {test_case_name}",
    ) as my_circ_ref:
        assert isinstance(my_circ_ref, CircuitRef)
        assert isinstance(my_circ_ref.download_circuit(), Circuit)

        my_circ_2 = qnx.circuits.get(id=my_circ_ref.id)

        # For some reason direct equality check fails
        assert my_circ_ref.id == my_circ_2.id
        assert my_circ_ref.annotations == my_circ_2.annotations
        assert my_circ_ref.project.id == my_circ_2.project.id

        with pytest.raises(qnx_exc.NoUniqueMatch):
            qnx.circuits.get()

        with pytest.raises(qnx_exc.ZeroMatches):
            qnx.circuits.get(name_like=f"{test_case_name}-wrong")

        test_ref_serialisation("circuit", my_circ_2)


def test_circuit_get_all(
    test_case_name: str,
    create_circuit_in_project: Callable[
        [Circuit, str, str], ContextManager[CircuitRef]
    ],
) -> None:
    """Test that we can get an iterator over all circuits."""

    project_name = f"project for {test_case_name}"

    with create_circuit_in_project(
        test_circuit,
        project_name,
        f"circuit1 for {test_case_name}",
    ):
        with create_circuit_in_project(
            test_circuit,
            project_name,
            f"circuit2 for {test_case_name}",
        ):
            my_circ_db_matches = qnx.circuits.get_all(name_like=test_case_name)

            assert my_circ_db_matches.count() == 2
            assert isinstance(my_circ_db_matches.summarize(), pd.DataFrame)
            assert isinstance(next(my_circ_db_matches), CircuitRef)


def test_circuit_create(
    test_case_name: str,
    create_property_in_project: Callable[..., ContextManager[ProjectRef]],
) -> None:
    """Test that we can create a circuit and add a property value."""

    project_name = f"project for {test_case_name}"
    property_name = f"property for {test_case_name}"

    with create_property_in_project(
        project_name=project_name,
        property_name=property_name,
        property_type="string",
        required=False,
    ):
        my_proj = qnx.projects.get(name_like=project_name)

        circuit_name = f"circuit for {test_case_name}"
        my_circ = Circuit(2, 2).H(0).CX(0, 1).measure_all()
        my_new_circuit = qnx.circuits.upload(
            circuit=my_circ, name=circuit_name, project=my_proj
        )

        assert isinstance(my_new_circuit, CircuitRef)

        test_prop_value = "foo"

        updated_circuit_ref = qnx.circuits.update(
            ref=my_new_circuit,
            properties=PropertiesDict({property_name: test_prop_value}),
        )

        assert (
            updated_circuit_ref.annotations.properties[property_name] == test_prop_value
        )


def test_circuit_get_cost(
    test_case_name: str,
    create_project: Callable[[str], ContextManager[ProjectRef]],
) -> None:
    """Test that we can get the cost to run a CircuitRef,
    on a particular Quantinuum Systems device."""

    with create_project(f"project for {test_case_name}") as my_proj:
        my_q_systems_circuit = qnx.circuits.upload(
            circuit=Circuit(2, 2).ZZPhase(0.5, 0, 1).measure_all(),
            name="qa_q_systems_circuit",
            project=my_proj,
        )

        cost = qnx.circuits.cost(
            circuit_ref=my_q_systems_circuit,
            n_shots=10,
            backend_config=qnx.QuantinuumConfig(device_name="H1-1E"),
            syntax_checker="H1-1SC",
        )

        assert isinstance(cost, float)
        assert cost > 0.0



"""Additional tests for references."""

from typing import get_args

from qnexus.models.references import BaseRef, Ref


def test_base_ref() -> None:
    """Test that all subclasses of BaseRef are the same
    as the generic Ref Union type."""

    all_base_refs = set(BaseRef.__subclasses__())

    all_refs = set(get_args(Ref.__args__[0]))  # type: ignore

    assert all_base_refs == all_refs



"""Test basic functionality relating to the team module."""

from typing import Callable, ContextManager

import pandas as pd

import qnexus as qnx
from qnexus.models.references import Ref, TeamRef


def test_team_get(
    test_case_name: str,
    create_team: Callable[[str], ContextManager[TeamRef]],
    test_ref_serialisation: Callable[[str, Ref], None],
) -> None:
    """Test that we can get a specific unique TeamRef and its
    serialisation round trip."""

    team_name = f"{test_case_name[-86:]}"  # TODO: use full name once bug is fixed

    with create_team(team_name):
        my_team = qnx.teams.get(name=team_name)
        assert isinstance(my_team, TeamRef)

        test_ref_serialisation("team", my_team)


def test_team_get_all(
    test_case_name: str,
    create_team: Callable[[str], ContextManager[TeamRef]],
) -> None:
    """Test that we can get all teams (currently not a NexusIterator)."""

    team_name = f"{test_case_name[-86:]}"  # TODO: use full name once bug is fixed

    with create_team(team_name):
        my_teams = qnx.teams.get_all()

        assert len(my_teams) >= 1
        assert isinstance(my_teams.df(), pd.DataFrame)
        assert isinstance(my_teams[0], TeamRef)


def test_team_create(
    test_case_name: str,
    authenticated_nexus: None,
) -> None:
    """Test that we can create a team."""

    team_name = f"{test_case_name[-86:]}"  # TODO: use full name once bug is fixed
    team_description = "A test team (can be deleted)"

    my_new_team = qnx.teams.create(name=team_name, description=team_description)

    assert isinstance(my_new_team, TeamRef)
    assert my_new_team.name == team_name
    assert my_new_team.description == team_description



"""Client API for jobs in Nexus."""

import asyncio
import json
import ssl
from datetime import datetime, timezone
from enum import Enum
from typing import Any, Type, Union, cast, overload
from uuid import UUID

import httpx
from quantinuum_schemas.models.backend_config import config_name_to_class
from quantinuum_schemas.models.hypertket_config import HyperTketConfig
from websockets.asyncio.client import connect, process_exception
from websockets.exceptions import ConnectionClosed

import qnexus.exceptions as qnx_exc
from qnexus.client import get_nexus_client
from qnexus.client.jobs import _compile, _execute
from qnexus.client.nexus_iterator import NexusIterator
from qnexus.client.utils import accept_circuits_for_programs, handle_fetch_errors
from qnexus.config import CONFIG
from qnexus.context import (
    get_active_project,
    merge_project_from_context,
    merge_properties_from_context,
)
from qnexus.models import BackendConfig
from qnexus.models.annotations import Annotations, PropertiesDict
from qnexus.models.filters import (
    CreatorFilter,
    FuzzyNameFilter,
    JobStatusFilter,
    JobTypeFilter,
    PaginationFilter,
    ProjectRefFilter,
    PropertiesFilter,
    ScopeFilter,
    ScopeFilterEnum,
    SortFilter,
    SortFilterEnum,
    TimeFilter,
)
from qnexus.models.job_status import WAITING_STATUS, JobStatus, JobStatusEnum
from qnexus.models.language import Language
from qnexus.models.references import (
    CircuitRef,
    CompilationResultRef,
    CompileJobRef,
    DataframableList,
    ExecuteJobRef,
    ExecutionProgram,
    ExecutionResult,
    ExecutionResultRef,
    GpuDecoderConfigRef,
    JobRef,
    JobType,
    ProjectRef,
    SystemRef,
    WasmModuleRef,
)
from qnexus.models.utils import assert_never

EPOCH_START = datetime(1970, 1, 1, tzinfo=timezone.utc)


class RemoteRetryStrategy(str, Enum):
    """Strategy to use when retrying jobs.

    Each strategy defines how the system should approach resolving
    potential conflicts with remote state.

    DEFAULT will only attempt to re-sync status and collect results
    from the third party. Duplicate results will not be saved.

    ALLOW_RESUBMIT will submit the job to the third party again if
    # the system has no record of a third party handle.

    FORCE_RESUBMIT will submit the job to the third party again if
    the system has a job handle already but no result.

    FULL_RESTART will act as though the job is entirely fresh and
    re-perform every action.
    """

    DEFAULT = "DEFAULT"
    ALLOW_RESUBMIT = "ALLOW_RESUBMIT"
    FORCE_RESUBMIT = "FORCE_RESUBMIT"
    FULL_RESTART = "FULL_RESTART"


class Params(
    CreatorFilter,
    PropertiesFilter,
    PaginationFilter,
    FuzzyNameFilter,
    JobStatusFilter,
    ProjectRefFilter,
    JobTypeFilter,
    ScopeFilter,
    SortFilter,
    TimeFilter,
):
    """Params for filtering jobs"""


@merge_project_from_context
def get_all(
    name_like: str | None = None,
    creator_email: list[str] | None = None,
    project: ProjectRef | None = None,
    properties: PropertiesDict | None = None,
    job_status: list[JobStatusEnum] | None = None,
    job_type: list[JobType] | None = None,
    created_before: datetime | None = None,
    created_after: datetime | None = datetime(day=1, month=1, year=2023),
    modified_before: datetime | None = None,
    modified_after: datetime | None = None,
    sort_filters: list[SortFilterEnum] | None = None,
    page_number: int | None = None,
    page_size: int | None = None,
    scope: ScopeFilterEnum | None = None,
) -> NexusIterator[CompileJobRef | ExecuteJobRef]:
    """Get a NexusIterator over jobs with optional filters."""
    project = project or get_active_project(project_required=False)
    project = cast(ProjectRef, project)

    params = Params(
        name_like=name_like,
        creator_email=creator_email,
        project=project,
        status=(
            JobStatusFilter.convert_status_filters(job_status) if job_status else None
        ),
        job_type=job_type,
        properties=properties,
        created_before=created_before,
        created_after=created_after,
        modified_before=modified_before,
        modified_after=modified_after,
        sort=SortFilter.convert_sort_filters(sort_filters),
        page_number=page_number,
        page_size=page_size,
        scope=scope,
    ).model_dump(by_alias=True, exclude_unset=True, exclude_none=True)

    return NexusIterator(
        resource_type="Job",
        nexus_url="/api/jobs/v1beta3",
        params=params,
        wrapper_method=_to_jobref,
        nexus_client=get_nexus_client(),
    )


def _to_jobref(data: dict[str, Any]) -> DataframableList[CompileJobRef | ExecuteJobRef]:
    """Parse a json dictionary into a list of JobRefs."""

    job_list: list[CompileJobRef | ExecuteJobRef] = []

    for entry in data["data"]:
        project_id = entry["relationships"]["project"]["data"]["id"]

        project_details = next(
            proj for proj in data["included"] if proj["id"] == project_id
        )
        project = ProjectRef(
            id=project_id,
            annotations=Annotations.from_dict(project_details["attributes"]),
            contents_modified=project_details["attributes"]["contents_modified"],
            archived=project_details["attributes"]["archived"],
        )

        system_id: str | None = (
            entry["relationships"]["system"]["data"]["id"]
            if "system" in entry["relationships"]
            else None
        )
        system_details = (
            next(item for item in data["included"] if item["id"] == system_id)
            if system_id is not None
            else None
        )

        system = (
            SystemRef(
                id=UUID(system_id),
                name=system_details["attributes"]["name"],
                provider_name=system_details["attributes"]["provider_name"],
            )
            if system_id is not None and system_details is not None
            else None
        )
        job_type: Type[CompileJobRef] | Type[ExecuteJobRef]
        match entry["attributes"]["job_type"]:
            case JobType.COMPILE:
                job_type = CompileJobRef
            case JobType.EXECUTE:
                job_type = ExecuteJobRef
            case _:
                assert_never(entry["attributes"]["job_type"])

        job_list.append(
            job_type(
                id=entry["id"],
                annotations=Annotations.from_dict(entry["attributes"]),
                job_type=entry["attributes"]["job_type"],
                last_status=JobStatus.from_dict(entry["attributes"]["status"]).status,
                last_message=JobStatus.from_dict(entry["attributes"]["status"]).message,
                project=project,
                system=system,
            )
        )
    return DataframableList(job_list)


def get(
    id: Union[str, UUID, None] = None,
    name_like: str | None = None,
    creator_email: list[str] | None = None,
    project: ProjectRef | None = None,
    properties: PropertiesDict | None = None,
    job_status: list[JobStatusEnum] | None = None,
    job_type: list[JobType] | None = None,
    created_before: datetime | None = None,
    created_after: datetime | None = datetime(day=1, month=1, year=2023),
    modified_before: datetime | None = None,
    modified_after: datetime | None = None,
    sort_filters: list[SortFilterEnum] | None = None,
    page_number: int | None = None,
    page_size: int | None = None,
    scope: ScopeFilterEnum | None = None,
) -> JobRef:
    """
    Get a single job using filters. Throws an exception if the filters do
    not match exactly one object.
    """
    if id:
        return _fetch_by_id(job_id=id, scope=scope)

    return get_all(
        name_like=name_like,
        creator_email=creator_email,
        project=project,
        properties=properties,
        job_status=job_status,
        job_type=job_type,
        created_before=created_before,
        created_after=created_after,
        modified_before=modified_before,
        modified_after=modified_after,
        sort_filters=sort_filters,
        page_number=page_number,
        page_size=page_size,
        scope=scope,
    ).try_unique_match()


def _fetch_by_id(job_id: UUID | str, scope: ScopeFilterEnum | None) -> JobRef:
    """Utility method for fetching directly by a unique identifier."""
    params = Params(
        scope=scope,
    ).model_dump(by_alias=True, exclude_unset=True, exclude_none=True)

    res = get_nexus_client().get(f"/api/jobs/v1beta3/{job_id}", params=params)

    handle_fetch_errors(res)

    job_data = res.json()

    project_id = job_data["data"]["relationships"]["project"]["data"]["id"]
    project_details = next(
        proj for proj in job_data["included"] if proj["id"] == project_id
    )
    project = ProjectRef(
        id=project_id,
        annotations=Annotations.from_dict(project_details["attributes"]),
        contents_modified=project_details["attributes"]["contents_modified"],
        archived=project_details["attributes"]["archived"],
    )

    system_id: str | None = (
        job_data["data"]["relationships"]["system"]["data"]["id"]
        if "system" in job_data["data"]["relationships"]
        else None
    )

    system_details = (
        next(item for item in job_data["included"] if item["id"] == system_id)
        if system_id is not None
        else None
    )

    system = (
        SystemRef(
            id=UUID(system_id),
            name=system_details["attributes"]["name"],
            provider_name=system_details["attributes"]["provider_name"],
        )
        if system_id is not None and system_details is not None
        else None
    )

    job_type: Type[CompileJobRef] | Type[ExecuteJobRef]
    match job_data["data"]["attributes"]["job_type"]:
        case JobType.COMPILE:
            job_type = CompileJobRef
        case JobType.EXECUTE:
            job_type = ExecuteJobRef
        case _:
            assert_never(job_data["attributes"]["job_type"])

    backend_config_dict = job_data["data"]["attributes"]["definition"]["backend_config"]
    backend_config_class = config_name_to_class[backend_config_dict["type"]]
    backend_config: BackendConfig = backend_config_class(  # type: ignore
        **backend_config_dict
    )

    return job_type(
        id=job_data["data"]["id"],
        annotations=Annotations.from_dict(job_data["data"]["attributes"]),
        job_type=job_data["data"]["attributes"]["job_type"],
        last_status=JobStatus.from_dict(
            job_data["data"]["attributes"]["status"]
        ).status,
        last_message=JobStatus.from_dict(
            job_data["data"]["attributes"]["status"]
        ).message,
        project=project,
        backend_config_store=backend_config,
        system=system,
    )


def wait_for(
    job: JobRef,
    wait_for_status: JobStatusEnum = JobStatusEnum.COMPLETED,
    timeout: float | None = 900.0,
) -> JobStatus:
    """Check job status until the job is complete (or a specified status)."""
    job_status = asyncio.run(
        asyncio.wait_for(
            listen_job_status(job=job, wait_for_status=wait_for_status),
            timeout=timeout,
        )
    )

    if (
        job_status.status == JobStatusEnum.ERROR
        and wait_for_status != JobStatusEnum.ERROR
    ):
        raise qnx_exc.JobError(f"Job errored with detail: {job_status.error_detail}")
    if (
        job_status.status == JobStatusEnum.CANCELLED
        and wait_for_status != JobStatusEnum.CANCELLED
    ):
        raise qnx_exc.JobError("Job was cancelled")
    if (
        job_status.status == JobStatusEnum.DEPLETED
        and wait_for_status != JobStatusEnum.DEPLETED
    ):
        raise qnx_exc.JobError("Job has run out of account credits")
    if (
        job_status.status == JobStatusEnum.TERMINATED
        and wait_for_status != JobStatusEnum.TERMINATED
    ):
        raise qnx_exc.JobError("Job has been terminated")

    return job_status


def status(job: JobRef) -> JobStatus:
    """Get the status of a job."""
    resp = get_nexus_client().get(f"api/jobs/v1beta3/{job.id}/attributes/status")
    if resp.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(
            message=resp.text, status_code=resp.status_code
        )
    job_status = JobStatus.from_dict(resp.json())
    # job.last_status = job_status.status
    return job_status


async def listen_job_status(
    job: JobRef, wait_for_status: JobStatusEnum = JobStatusEnum.COMPLETED
) -> JobStatus:
    """Check the Status of a Job via a websocket connection.
    Will use SSO tokens."""
    job_status = status(job)
    # logger.debug("Current job status: %s", job_status.status)
    if job_status.status not in WAITING_STATUS or job_status.status == wait_for_status:
        return job_status

    ssl_context = httpx.create_ssl_context(verify=CONFIG.httpx_verify)

    def _process_exception(exc: Exception) -> Exception | None:
        """Utility wrapper around process_exception that tells the websockets
        library not to auto-retry SSLErrors as they are usually not recoverable.

        Unfortunately SSLError inherits from OSError which websockets will always
        retried when `connect` is used in an async for loop.
        """
        if isinstance(exc, ssl.SSLError):
            return exc
        return process_exception(exc)

    additional_headers = {
        # TODO, this cookie will expire frequently
        "Cookie": f"myqos_id={get_nexus_client().auth.cookies.get('myqos_id')}"  # type: ignore
    }
    async for websocket in connect(
        f"{CONFIG.websockets_url}/api/jobs/v1beta3/{job.id}/attributes/status/ws",
        ssl=ssl_context,
        additional_headers=additional_headers,
        process_exception=_process_exception,
        # logger=logger,
    ):
        try:
            async for status_json in websocket:
                # logger.debug("New status: %s", status_json)
                job_status = JobStatus.from_dict(json.loads(status_json))

                if (
                    job_status.status not in WAITING_STATUS
                    or job_status.status == wait_for_status
                ):
                    break
            break
        except ConnectionClosed:
            # logger.debug(
            #     "Websocket connection closed... attempting to reconnect..."
            # )
            continue
        finally:
            try:
                await websocket.close(code=1000, reason="Client closed connection")
            except GeneratorExit:
                pass

    return job_status


@overload
def results(
    job: CompileJobRef, allow_incomplete: bool = False
) -> DataframableList[CompilationResultRef]: ...


@overload
def results(
    job: ExecuteJobRef, allow_incomplete: bool = False
) -> DataframableList[ExecutionResultRef]: ...


def results(
    job: CompileJobRef | ExecuteJobRef,
    allow_incomplete: bool = False,
) -> DataframableList[CompilationResultRef] | DataframableList[ExecutionResultRef]:
    """Get the ResultRefs from a JobRef, if the job is complete.
    To enable fetching results from Jobs with incomplete items, set allow_incomplete=True.
    """
    match job:
        case CompileJobRef():
            return _compile._results(job, allow_incomplete)
        case ExecuteJobRef():
            return _execute._results(job, allow_incomplete)
        case _:
            assert_never(job.job_type)


def retry_submission(
    job: JobRef,
    retry_status: list[JobStatusEnum] | None = None,
    remote_retry_strategy: RemoteRetryStrategy = RemoteRetryStrategy.DEFAULT,
    user_group: str | None = None,
) -> None:
    """Retry a job in Nexus according to status(es) or retry strategy.

    By default, jobs with the ERROR status will be retried.
    """
    body: dict[str, str | list[str]] = {"remote_retry_strategy": remote_retry_strategy}

    if user_group is not None:
        body["user_group"] = user_group

    if retry_status is not None:
        body["retry_status"] = [status.name for status in retry_status]

    res = get_nexus_client().post(
        f"/api/jobs/v1beta3/{job.id}/rpc/retry",
        json=body,
    )
    if res.status_code != 202:
        res.raise_for_status()


def cancel(job: JobRef) -> None:
    """Attempt cancellation of a job in Nexus.

    If the job has been submitted to a backend, Nexus will request cancellation of the job.
    """
    res = get_nexus_client().post(
        f"/api/jobs/v1beta3/{job.id}/rpc/cancel",
        json={},
    )

    if res.status_code != 202:
        res.raise_for_status()


def delete(job: JobRef) -> None:
    """Delete a job in Nexus."""
    res = get_nexus_client().delete(
        f"/api/jobs/v1beta3/{job.id}",
    )

    if res.status_code != 204:
        res.raise_for_status()


@accept_circuits_for_programs
@merge_properties_from_context
def compile(
    programs: Union[CircuitRef, list[CircuitRef]],
    backend_config: BackendConfig,
    name: str,
    description: str = "",
    project: ProjectRef | None = None,
    properties: PropertiesDict | None = None,
    optimisation_level: int = 2,
    credential_name: str | None = None,
    user_group: str | None = None,
    hypertket_config: HyperTketConfig | None = None,
    timeout: float | None = 300.0,
) -> DataframableList[CircuitRef]:
    """
    Utility method to run a compile job on a program or programs and return a
    DataframableList of the compiled programs.
    """
    project = project or get_active_project(project_required=True)
    project = cast(ProjectRef, project)

    compile_job_ref = _compile.start_compile_job(
        programs=programs,
        backend_config=backend_config,
        name=name,
        description=description,
        project=project,
        properties=properties,
        optimisation_level=optimisation_level,
        credential_name=credential_name,
        user_group=user_group,
        hypertket_config=hypertket_config,
    )

    wait_for(job=compile_job_ref, timeout=timeout)

    compile_results = results(compile_job_ref)

    return DataframableList(
        [compile_result.get_output() for compile_result in compile_results]
    )


@accept_circuits_for_programs
@merge_properties_from_context
def execute(
    programs: Union[ExecutionProgram, list[ExecutionProgram]],
    n_shots: list[int] | list[None],
    backend_config: BackendConfig,
    name: str,
    description: str = "",
    properties: PropertiesDict | None = None,
    project: ProjectRef | None = None,
    valid_check: bool = True,
    wasm_module: WasmModuleRef | None = None,
    gpu_decoder_config: GpuDecoderConfigRef | None = None,
    language: Language = Language.AUTO,
    credential_name: str | None = None,
    user_group: str | None = None,
    timeout: float | None = 300.0,
) -> list[ExecutionResult]:
    """
    Utility method to run an execute job and return the results. Blocks until
    the results are available. See ``qnexus.start_execute_job`` for a function
    that submits the job and returns immediately, rather than waiting for
    results.
    """

    execute_job_ref = _execute.start_execute_job(
        programs=programs,
        n_shots=n_shots,
        backend_config=backend_config,
        name=name,
        description=description,
        properties=properties,
        project=project,
        valid_check=valid_check,
        wasm_module=wasm_module,
        gpu_decoder_config=gpu_decoder_config,
        language=language,
        credential_name=credential_name,
        user_group=user_group,
    )

    wait_for(job=execute_job_ref, timeout=timeout)

    execute_results = results(execute_job_ref)

    return [result.download_result() for result in execute_results]


def cost(job: CompileJobRef | ExecuteJobRef) -> float:
    """Get the HQC cost of a job from a JobRef."""
    resp = get_nexus_client().get(f"/api/jobs/v1beta3/{job.id}")
    if resp.status_code != 200:
        raise qnx_exc.ResourceFetchFailed(
            message=resp.text, status_code=resp.status_code
        )
    resp_data = resp.json()["data"]
    job_status = resp_data["attributes"]["status"]
    cost = job_status.get("cost")
    return float(cost) if cost is not None else 0.0



"""Test basic functionality relating to the job module."""

from collections import Counter
from datetime import datetime
from time import sleep
from typing import Any, Callable, ContextManager

import pandas as pd
import pytest
from pytket.backends.backendinfo import BackendInfo
from pytket.backends.backendresult import BackendResult
from pytket.circuit import Circuit
from quantinuum_schemas.models.backend_config import BaseBackendConfig
from quantinuum_schemas.models.hypertket_config import HyperTketConfig

import qnexus as qnx
import qnexus.exceptions as qnx_exc
from qnexus.models.job_status import JobStatusEnum
from qnexus.models.references import (
    CircuitRef,
    CompilationPassRef,
    CompilationResultRef,
    CompileJobRef,
    ExecuteJobRef,
    JobRef,
    Ref,
)

# The following global variables and autoused fixture are a
# bit of a hack to have global identifiers for the resources
# used by the `*job_get` tests in this suite. Using the same name for the
# resources means they will be reused if they exist,
# as the "create_*_in_project" fixtures do precisely that.
project_name = "project for {test_suite_name}"
circuit_name = "circuit for {test_suite_name}"
compile_job_name = "compile job for {test_suite_name}"
execute_job_name = "execute job for {test_suite_name}"


@pytest.fixture(autouse=True)
def set_resource_names(test_suite_name: str) -> None:
    global project_name
    global circuit_name
    global compile_job_name
    global execute_job_name

    project_name = project_name.replace("{test_suite_name}", test_suite_name)
    circuit_name = circuit_name.replace("{test_suite_name}", test_suite_name)
    compile_job_name = compile_job_name.replace("{test_suite_name}", test_suite_name)
    execute_job_name = execute_job_name.replace("{test_suite_name}", test_suite_name)


def test_job_get_all(
    create_compile_job_in_project: Callable[..., ContextManager[CompileJobRef]],
    create_execute_job_in_project: Callable[..., ContextManager[ExecuteJobRef]],
    test_circuit: Circuit,
) -> None:
    """Test that we can get an iterator over all jobs."""

    with create_compile_job_in_project(
        project_name=project_name,
        job_name=compile_job_name,
        circuit=test_circuit,
        circuit_name=circuit_name,
    ):
        with create_execute_job_in_project(
            project_name=project_name,
            job_name=execute_job_name,
            circuit=test_circuit,
            circuit_name=circuit_name,
        ):
            proj_ref = qnx.projects.get(name_like=project_name)
            my_job_db_matches = qnx.jobs.get_all(project=proj_ref)

            assert isinstance(my_job_db_matches.count(), int)
            assert isinstance(my_job_db_matches.summarize(), pd.DataFrame)

            for job_ref in my_job_db_matches.list():
                assert isinstance(job_ref, JobRef)
                assert job_ref.backend_config is not None
                assert isinstance(job_ref.backend_config, BaseBackendConfig)


def test_job_get_by_id(
    create_compile_job_in_project: Callable[..., ContextManager[CompileJobRef]],
    create_execute_job_in_project: Callable[..., ContextManager[ExecuteJobRef]],
    test_circuit: Circuit,
    test_ref_serialisation: Callable[[str, Ref], None],
) -> None:
    """Test that we can get JobRefs by id and their serialisation."""

    def get_sanitised_job_ref(job_ref: CompileJobRef | ExecuteJobRef) -> dict[str, Any]:
        """Return the model_dump of a JobRef but excluding the fields that is ok if they
        have changed."""
        return job_ref.model_dump(
            exclude={
                "last_message": True,
                "last_status": True,
                "system": True,
                "__all__": {"contents_modified": True, "modified": True},
            }
        )

    with create_compile_job_in_project(
        project_name=project_name,
        job_name=compile_job_name,
        circuit=test_circuit,
        circuit_name=circuit_name,
    ) as compile_job_ref:
        with create_execute_job_in_project(
            project_name=project_name,
            job_name=execute_job_name,
            circuit=test_circuit,
            circuit_name=circuit_name,
        ) as execute_job_ref:
            assert isinstance(compile_job_ref, CompileJobRef)
            assert isinstance(compile_job_ref.backend_config, BaseBackendConfig)

            my_compile_job_by_id = qnx.jobs.get(id=compile_job_ref.id)
            assert isinstance(my_compile_job_by_id, CompileJobRef)

            # Some attributes might have changed, so need to mask them out while comparing
            assert get_sanitised_job_ref(compile_job_ref) == get_sanitised_job_ref(
                my_compile_job_by_id
            )
            test_ref_serialisation("compile", my_compile_job_by_id)

            assert isinstance(execute_job_ref, ExecuteJobRef)
            assert isinstance(execute_job_ref.backend_config, BaseBackendConfig)

            my_execute_job_by_id = qnx.jobs.get(id=execute_job_ref.id)
            assert isinstance(my_execute_job_by_id, ExecuteJobRef)

            # Some attributes might have changed, so need to mask them out while comparing
            assert get_sanitised_job_ref(execute_job_ref) == get_sanitised_job_ref(
                my_execute_job_by_id
            )

            test_ref_serialisation("execute", my_execute_job_by_id)


def test_compile_job_get(
    create_compile_job_in_project: Callable[..., ContextManager[CompileJobRef]],
    test_circuit: Circuit,
) -> None:
    """Test that we can get a specific unique CompileJobRef,
    or get an appropriate exception."""

    with create_compile_job_in_project(
        project_name=project_name,
        job_name=compile_job_name,
        circuit=test_circuit,
        circuit_name=circuit_name,
    ) as compile_job_ref:
        assert isinstance(compile_job_ref, CompileJobRef)

        with pytest.raises(qnx_exc.NoUniqueMatch):
            qnx.jobs.get()

        with pytest.raises(qnx_exc.ZeroMatches):
            qnx.jobs.get(name_like=f"{datetime.now()}_{datetime.now()}")


def test_execute_job_get(
    create_execute_job_in_project: Callable[..., ContextManager[ExecuteJobRef]],
    test_circuit: Circuit,
) -> None:
    """Test that we can get a specific unique ExecuteJobRef,
    or get an appropriate exception."""
    with create_execute_job_in_project(
        project_name=project_name,
        job_name=execute_job_name,
        circuit=test_circuit,
        circuit_name=circuit_name,
    ) as execute_job_ref:
        assert isinstance(execute_job_ref, ExecuteJobRef)

        with pytest.raises(qnx_exc.NoUniqueMatch):
            qnx.jobs.get()

        with pytest.raises(qnx_exc.ZeroMatches):
            qnx.jobs.get(name_like=f"{datetime.now()}_{datetime.now()}")


def test_submit_compile(
    create_compile_job_in_project: Callable[..., ContextManager[CompileJobRef]],
    test_circuit: Circuit,
    test_ref_serialisation: Callable[[str, Ref], None],
) -> None:
    """Test that we can run a compile job in Nexus, wait for the job to complete and
    obtain the results from the compilation."""
    config = qnx.AerConfig()

    with create_compile_job_in_project(
        project_name=project_name,
        job_name=compile_job_name,
        circuit=test_circuit,
        circuit_name=circuit_name,
        backend_config=config,
    ) as compile_job_ref:
        assert isinstance(compile_job_ref, CompileJobRef)

        qnx.jobs.wait_for(compile_job_ref)

        compile_results = qnx.jobs.results(compile_job_ref)

        assert len(compile_results) == 1
        assert isinstance(compile_results[0], CompilationResultRef)
        test_ref_serialisation("compile_result", compile_results[0])

        assert isinstance(compile_results[0].get_input(), CircuitRef)
        assert isinstance(compile_results[0].get_output(), CircuitRef)

        first_pass_data = compile_results[0].get_passes()[0]

        assert isinstance(first_pass_data, CompilationPassRef)
        assert isinstance(first_pass_data.get_input(), CircuitRef)
        assert isinstance(first_pass_data.get_output(), CircuitRef)
        assert isinstance(first_pass_data.pass_name, str)
        test_ref_serialisation("compilation_pass", first_pass_data)

        cj_ref = qnx.jobs.get(id=compile_job_ref.id)
        assert cj_ref.backend_config == config


def test_compile(
    test_case_name: str,
    create_circuit_in_project: Callable[
        [Circuit, str, str], ContextManager[CircuitRef]
    ],
    test_circuit: Circuit,
) -> None:
    """Test that we can run the utility compile function and get compiled circuits."""

    local_project_name = f"project for {test_case_name}"
    local_circuit_name = f"circuit for {test_case_name}"
    local_compile_job_name = f"compile job for {test_case_name}"

    with create_circuit_in_project(
        test_circuit,
        local_project_name,
        local_circuit_name,
    ) as circ_ref:
        my_proj = qnx.projects.get(name_like=local_project_name)

        compiled_circuits = qnx.compile(
            programs=[circ_ref],
            name=local_compile_job_name,
            project=my_proj,
            backend_config=qnx.AerConfig(),
        )

        assert len(compiled_circuits) == 1
        assert isinstance(compiled_circuits[0], CircuitRef)


def test_get_results_for_incomplete_compile(
    test_case_name: str,
    create_circuit_in_project: Callable[
        [Circuit, str, str], ContextManager[CircuitRef]
    ],
    test_circuit: Circuit,
) -> None:
    """Test that we can run a compile job in Nexus, and fetch complete results for
    an otherwise errored/incomplete job."""

    local_project_name = f"project for {test_case_name}"
    local_circuit_name = f"circuit for {test_case_name}"

    with create_circuit_in_project(
        test_circuit,
        local_project_name,
        local_circuit_name,
    ) as circ_ref:
        proj_ref = qnx.projects.get_or_create(local_project_name)

        # Will create `n_jobs` compile jobs to guarantee that the last job
        # has not completed before checking the incomplete results
        n_jobs = 10
        compile_job_ref = None
        for job_num in range(n_jobs):
            compile_job_name = f"compile job {job_num} for {test_case_name}"

            compile_job_ref = qnx.start_compile_job(
                programs=[circ_ref],
                name=compile_job_name,
                project=proj_ref,
                backend_config=qnx.AerConfig(),
            )

        assert isinstance(compile_job_ref, CompileJobRef)
        assert qnx.jobs.status(compile_job_ref).status != JobStatusEnum.COMPLETED
        compile_results = qnx.jobs.results(compile_job_ref, allow_incomplete=True)
        assert len(compile_results) == 0


def test_compile_hypertket(
    test_case_name: str,
    create_circuit_in_project: Callable[
        [Circuit, str, str], ContextManager[CircuitRef]
    ],
    test_circuit: Circuit,
) -> None:
    """Test that we can run compile circuits for hypertket."""

    local_project_name = f"project for {test_case_name}"
    local_circuit_name = f"circuit for {test_case_name}"
    local_job_name = f"hypertket job for {test_case_name}"

    with create_circuit_in_project(
        test_circuit,
        local_project_name,
        local_circuit_name,
    ) as circ_ref:
        my_proj = qnx.projects.get(name_like=local_project_name)

        compiled_circuits = qnx.compile(
            programs=[circ_ref],
            name=local_job_name,
            project=my_proj,
            backend_config=qnx.QuantinuumConfig(device_name="H1-1LE"),
            hypertket_config=HyperTketConfig(),
        )

        assert len(compiled_circuits) == 1
        assert isinstance(compiled_circuits[0], CircuitRef)


def test_submit_execute(
    create_execute_job_in_project: Callable[..., ContextManager[ExecuteJobRef]],
    test_circuit: Circuit,
    # test_ref_serialisation: Callable[[str, Ref], None],
) -> None:
    """Test that we can run an execute job in Nexus, wait for the job to complete and
    obtain the results from the execution."""
    config = qnx.AerConfig()

    with create_execute_job_in_project(
        project_name=project_name,
        job_name=execute_job_name,
        circuit=test_circuit,
        circuit_name=circuit_name,
    ) as execute_job_ref:
        qnx.jobs.wait_for(execute_job_ref)

        execute_results = qnx.jobs.results(execute_job_ref)

        assert len(execute_results) == 1

        first_result = execute_results[0]
        assert isinstance(first_result.get_input(), CircuitRef)
        assert isinstance(first_result.download_result(), BackendResult)
        assert isinstance(first_result.download_backend_info(), BackendInfo)

        pj_ref = qnx.jobs.get(id=execute_job_ref.id)
        assert pj_ref.backend_config == config

        # TODO: check why the serialisation round trip fails even when the
        # objects are the same
        # test_ref_serialisation(ref_type="execute_result", ref=first_result)


def test_execute(
    test_case_name: str,
    create_circuit_in_project: Callable[
        [Circuit, str, str], ContextManager[CircuitRef]
    ],
    test_circuit: Circuit,
) -> None:
    """Test that we can run the utility execute function and get the results of the execution."""
    local_project_name = f"project for {test_case_name}"
    local_circuit_name = f"circuit for {test_case_name}"
    local_execute_job_name = f"execute job for {test_case_name}"

    with create_circuit_in_project(
        test_circuit,
        local_project_name,
        local_circuit_name,
    ) as circ_ref:
        my_proj = qnx.projects.get_or_create(local_project_name)

        backend_results = qnx.execute(
            programs=[circ_ref],
            name=local_execute_job_name,
            project=my_proj,
            backend_config=qnx.AerConfig(),
            n_shots=[10],
        )

        assert len(backend_results) == 1
        assert isinstance(backend_results[0], BackendResult)
        assert isinstance(backend_results[0].get_counts(), Counter)


def test_get_results_for_incomplete_execute(
    test_case_name: str,
    create_circuit_in_project: Callable[
        [Circuit, str, str], ContextManager[CircuitRef]
    ],
    test_circuit: Circuit,
) -> None:
    """Test that we can run an execute job in Nexus, and fetch complete results for
    an otherwise errored/incomplete job."""

    local_project_name = f"project for {test_case_name}"
    local_circuit_name = f"circuit for {test_case_name}"
    job_name = f"execute job for {test_case_name}"

    with create_circuit_in_project(
        test_circuit,
        local_project_name,
        local_circuit_name,
    ) as circ_ref:
        proj_ref = qnx.projects.get_or_create(local_project_name)

        my_q_systems_circuit = qnx.circuits.upload(
            circuit=Circuit(2, 2).ZZPhase(0.5, 0, 1).measure_all(),
            name="qa_q_systems_circuit",
            project=proj_ref,
        )

        execute_job_ref = qnx.start_execute_job(
            programs=[circ_ref, my_q_systems_circuit],
            name=job_name,
            project=proj_ref,
            backend_config=qnx.QuantinuumConfig(device_name="H1-1LE"),
            n_shots=[10, 10],
        )

        assert isinstance(execute_job_ref, ExecuteJobRef)

        with pytest.raises(qnx_exc.JobError):
            qnx.jobs.wait_for(execute_job_ref)

        incomplete_results = qnx.jobs.results(execute_job_ref, allow_incomplete=True)

        # wait for the ZZPhase circuit execution to complete
        for _ in range(10):
            incomplete_results = qnx.jobs.results(
                execute_job_ref, allow_incomplete=True
            )
            if len(incomplete_results) > 0:
                break
            sleep(10)

        # we expect the CX circuit to fail on H1-1LE, but the ZZPhase circuit should succeed
        assert len(incomplete_results) == 1

        assert isinstance(incomplete_results[0].get_input(), CircuitRef)
        assert isinstance(incomplete_results[0].download_result(), BackendResult)
        assert isinstance(incomplete_results[0].download_backend_info(), BackendInfo)


def test_wait_for_raises_on_job_error(
    test_case_name: str,
    create_circuit_in_project: Callable[
        [Circuit, str, str], ContextManager[CircuitRef]
    ],
    test_circuit: Circuit,
) -> None:
    """Test that if a job errors, the wait_for function raises an Exception."""

    local_project_name = f"project for {test_case_name}"
    local_circuit_name = f"circuit for {test_case_name}"
    job_name = f"execute job for {test_case_name}"

    with create_circuit_in_project(
        test_circuit,
        local_project_name,
        local_circuit_name,
    ) as circ_ref:
        proj_ref = qnx.projects.get_or_create(local_project_name)

        # Circuit not compiled for H1-1LE, so we expect it to error when executed
        failing_job_ref = qnx.start_execute_job(
            programs=[circ_ref],
            name=job_name,
            project=proj_ref,
            n_shots=[10],
            backend_config=qnx.QuantinuumConfig(device_name="H1-1LE"),
        )

        with pytest.raises(qnx_exc.ResourceFetchFailed):
            qnx.jobs.results(failing_job_ref)

        with pytest.raises(qnx_exc.JobError):
            qnx.jobs.wait_for(failing_job_ref)


def test_results_not_available_error(
    test_case_name: str,
    create_circuit_in_project: Callable[
        [Circuit, str, str], ContextManager[CircuitRef]
    ],
    test_circuit: Circuit,
) -> None:
    """Test that we can't get the results of a job until it is complete."""

    local_project_name = f"project for {test_case_name}"
    local_circuit_name = f"circuit for {test_case_name}"
    job_name = f"execute job for {test_case_name}"

    with create_circuit_in_project(
        test_circuit,
        local_project_name,
        local_circuit_name,
    ) as circ_ref:
        proj_ref = qnx.projects.get_or_create(local_project_name)

        execute_job_ref = qnx.start_execute_job(
            programs=[circ_ref],
            name=job_name,
            project=proj_ref,
            backend_config=qnx.AerConfig(),
            n_shots=[10],
        )

        with pytest.raises(qnx_exc.ResourceFetchFailed):
            qnx.jobs.results(execute_job_ref)

        qnx.jobs.wait_for(execute_job_ref)

        execute_results = qnx.jobs.results(execute_job_ref)

        assert len(execute_results) == 1

        assert isinstance(execute_results[0].get_input(), CircuitRef)

        assert isinstance(execute_results[0].download_result(), BackendResult)

        assert isinstance(execute_results[0].download_backend_info(), BackendInfo)


def test_submit_under_user_group(
    test_case_name: str,
    create_circuit_in_project: Callable[
        [Circuit, str, str], ContextManager[CircuitRef]
    ],
    test_circuit: Circuit,
) -> None:
    """Test that a user can submit jobs under a user_group that
    they belong to.

    Requires that the test user is a member of a group called:
    'QA_IntegrationTestGroup',
    and not a member of a group called:
    'made_up_group'.
    """

    local_project_name = f"project for {test_case_name}"
    local_circuit_name = f"circuit for {test_case_name}"
    correct_group = "QA_IntegrationTestGroup"
    fake_group = "made_up_group"

    with create_circuit_in_project(
        test_circuit,
        local_project_name,
        local_circuit_name,
    ) as circ_ref:
        my_proj = qnx.projects.get_or_create(local_project_name)

        with pytest.raises(qnx_exc.ResourceCreateFailed) as exc:
            qnx.start_compile_job(
                programs=[circ_ref],
                name=f"compile job xfail for {test_case_name}",
                project=my_proj,
                backend_config=qnx.AerConfig(),
                user_group=fake_group,
            )
            assert (
                exc.value.message
                == f"Not a member of any group with name: {fake_group}"
            )

        qnx.start_compile_job(
            programs=[circ_ref],
            name=f"compile job for {test_case_name}",
            project=my_proj,
            backend_config=qnx.AerConfig(),
            user_group=correct_group,
        )

        with pytest.raises(qnx_exc.ResourceCreateFailed):
            qnx.start_execute_job(
                programs=[circ_ref],
                name=f"execute job xfail for {test_case_name}",
                project=my_proj,
                backend_config=qnx.AerConfig(),
                n_shots=[10],
                user_group=fake_group,
            )
            assert (
                exc.value.message
                == f"Not a member of any group with name: {fake_group}"
            )

        qnx.start_execute_job(
            programs=[circ_ref],
            name=f"execute job for {test_case_name}",
            project=my_proj,
            backend_config=qnx.AerConfig(),
            n_shots=[10],
            user_group=correct_group,
        )
